#!/usr/bin/python3
# mypy: disable-error-code="index"

from typing import Union

from semantic_version import Version

from brownie._c_constants import Path
from brownie._config import _get_data_folder
from brownie.typing import ContractName, Source

VersionSpec = Union[str, Version]
VersionList = list[Version]


def expand_source_map(source_map_str: str | dict) -> list[Source]:
    """Expand the compressed sourceMap supplied by solc into a list of Tuple[Start, Stop, ContractName, str]."""

    if isinstance(source_map_str, dict):
        # NOTE: vyper >= 0.4 gives us a dict that contains the source map
        source_map_str = source_map_str["pc_pos_map_compressed"]
    if not isinstance(source_map_str, str):
        raise TypeError(source_map_str)

    source_map = [_expand_row(i) if i else None for i in source_map_str.split(";")]
    for i, value in enumerate(source_map[1:], 1):
        if value is None:
            source_map[i] = source_map[i - 1]
            continue
        for x in range(4):
            if value[x] is None:
                value[x] = source_map[i - 1][x]
    return [tuple(_list) for _list in source_map]  # type: ignore [arg-type, misc]


def _expand_row(row: str) -> list[str | int | None]:
    """Expand a packed string into a row of params."""
    result: list[str | int | None] = [None] * 4
    # ignore the new "modifier depth" value in solidity 0.6.0
    for i, value in enumerate(row.split(":")[:4]):
        if value:
            result[i] = value if i == 3 else int(value)
    return result


def merge_natspec(
    devdoc: dict[str, dict[str, dict]],
    userdoc: dict[str, dict[str, dict]],
) -> dict[str, dict[str, dict]]:
    """
    Merge devdoc and userdoc compiler output to a single dict.

    Arguments
    ---------
    devdoc: dict
        Devdoc compiler output.
    userdoc : dict
        Userdoc compiler output.

    Returns
    -------
    dict
        Combined natspec.
    """
    natspec = {"methods": {}, **userdoc, **devdoc}
    usermethods = userdoc.get("methods", {})
    devmethods = devdoc.get("methods", {})

    keys: set[str] = set()
    keys.update(usermethods)
    keys.update(devmethods)
    for key in keys:
        try:
            natspec["methods"][key] = {**usermethods.get(key, {}), **devmethods.get(key, {})}
        except TypeError:
            # sometimes Solidity has inconsistent NatSpec formatting ¯\_(ツ)_/¯
            pass
    return natspec


def _get_alias(contract_name: ContractName, path_str: str) -> ContractName:
    # Generate an alias for a contract, used when tracking dependencies.
    # For a contract within the project, the alias == the name. For contracts
    # imported from a dependency, the alias is set as [PACKAGE]/[NAME]
    # to avoid namespace collisions.
    data_path = _get_data_folder().parts
    path_parts = Path(path_str).parts
    if path_parts[: len(data_path)] == data_path:
        idx = len(data_path) + 1
        return f"{path_parts[idx]}/{path_parts[idx+1]}/{contract_name}"  # type: ignore [return-value]
    else:
        return contract_name



#!/usr/bin/python3

import sys
import traceback
from collections.abc import Sequence
from typing import Any, Final, Literal, cast, final

import pygments
from pygments.formatter import Formatter
from pygments.formatters import get_formatter_by_name
from pygments.lexers import PythonLexer
from pygments_lexer_solidity import SolidityLexer
from vyper.exceptions import VyperException

from brownie._c_constants import Path
from brownie._config import CONFIG

if sys.platform == "win32":
    import colorama

    colorama.init()

fmt_name = "terminal"
try:
    import curses

    curses.setupterm()
    if curses.tigetnum("colors") == 256:
        fmt_name = "terminal256"
except Exception:
    # if curses won't import we are probably using Windows
    pass

COLOR_STYLE: str = CONFIG.settings["console"]["color_style"]
formatter: Final[Formatter] = get_formatter_by_name(fmt_name, style=COLOR_STYLE)

if fmt_name == "terminal256" and COLOR_STYLE == "monokai":
    # dirty hack to make tree diagrams not have a fixed black blackground
    formatter.style_string["Token.Error"] = ("\x1b[0;2;37m", "\x1b[0;m")

BASE: Final = "\x1b[0;"

MODIFIERS: Final = {"bright": "1;", "dark": "2;"}

COLORS: Final = {
    "black": "30",
    "red": "31",
    "green": "32",
    "yellow": "33",
    "blue": "34",
    "magenta": "35",
    "cyan": "36",
    "white": "37",
}


NotifyType = Literal["SUCCESS", "WARNING", "ERROR"]
NOTIFY_COLORS: Final = {"WARNING": "bright red", "ERROR": "bright red", "SUCCESS": "bright green"}

base_path: Final = str(Path(".").absolute())


# cached color strings

blue: Final = "\x1b[0;34m"  # Color()("blue")
bright_black: Final = "\x1b[0;1;30m"  # Color()("bright black")
bright_blue: Final = "\x1b[0;1;34m"  # Color()("bright blue")
bright_cyan: Final = "\x1b[0;1;36m"  # Color()("bright cyan")
bright_green: Final = "\x1b[0;1;32m"  # Color()("bright green")
bright_magenta: Final = "\x1b[0;1;35m"  # Color()("bright magenta")
bright_red: Final = "\x1b[0;1;31m"  # Color()("bright red")
bright_yellow: Final = "\x1b[0;1;33m"  # Color()("bright yellow")
dark_white: Final = "\x1b[0;2;37m"  # Color()("dark white")
green: Final = "\x1b[0;1;32m"  # Color()("green")
red: Final = "\x1b[0;1;31m"  # Color()("red")
yellow: Final = "\x1b[0;1;33m"  # Color()("yellow")


@final
class Color:
    __cache__: Final[dict[str | None, str]] = {}

    def __call__(self, color_str: str | None = None) -> str:
        if not CONFIG.settings["console"]["show_colors"]:
            return ""
        try:
            return Color.__cache__[color_str]
        except KeyError:
            if not color_str:
                return f"{BASE}m"
            try:
                if " " in color_str:
                    modifier, color_str = color_str.split()
                    color = f"{BASE}{MODIFIERS[modifier]}{COLORS[color_str]}m"
                else:
                    color = f"{BASE}{COLORS[color_str]}m"
            except (KeyError, ValueError):
                color = f"{BASE}m"
            Color.__cache__[color_str] = color
            return color

    def __str__(self):
        return f"{BASE}m"

    # format dicts for console printing
    def pretty_dict(self, value: dict, _indent: int = 0) -> str:
        text = ""
        if not _indent:
            text = "{"
        _indent += 4
        for c, k in enumerate(sorted(value.keys(), key=str)):
            v = value[k]
            if c:
                text += ","
            s = "'" if isinstance(k, str) else ""
            text += f"\n{' '*_indent}{s}{k}{s}: "
            if isinstance(v, dict):
                text += "{" + self.pretty_dict(v, _indent)
                continue
            if isinstance(v, (list, tuple, set)):
                text += str(v)[0] + self.pretty_sequence(v, _indent)  # type: ignore [arg-type]
                continue
            text += self._write(v)
        _indent -= 4
        text += f"\n{' '*_indent}}}"
        return text

    # format lists for console printing
    def pretty_sequence(self, value: Sequence, _indent: int = 0) -> str:
        text = ""
        string = str(value)
        start_bracket, stop_bracket = string[0], string[-1]
        if not _indent:
            text += f"{start_bracket}"
        if value and not [i for i in value if not isinstance(i, dict)]:
            # list of dicts
            text += f"\n{' '*(_indent+4)}{{"
            text += f",\n{' '*(_indent+4)}{{".join(self.pretty_dict(i, _indent + 4) for i in value)
            text += f"\n{' '*_indent}{stop_bracket}"
        elif value and not [i for i in value if not isinstance(i, str) or len(i) != 64]:
            # list of bytes32 hexstrings (stack trace)
            text += ", ".join(f"\n{' '*(_indent+4)}{self._write(i)}" for i in value)
            text += f"\n{' '*_indent}{stop_bracket}"
        else:
            # all other cases
            text += ", ".join(self._write(i) for i in value)
            text += stop_bracket
        return text

    def _write(self, value: Any) -> str:
        s = '"' if isinstance(value, str) else ""
        return f"{s}{value}{s}"

    def format_tb(
        self,
        exc: BaseException,
        filename: str | None = None,
        start: int | None = None,
        stop: int | None = None,
    ) -> str:
        if isinstance(exc, SyntaxError) and exc.text is not None:
            return self.format_syntaxerror(exc)

        tb = [i.replace("./", "") for i in traceback.format_tb(exc.__traceback__)]
        if filename and not CONFIG.argv["tb"]:
            try:
                start = tb.index(next(i for i in tb if filename in i))
                stop = tb.index(next(i for i in tb[::-1] if filename in i)) + 1
            except Exception:
                pass

        tb = tb[start:stop]
        for i in range(len(tb)):
            info, code = tb[i].split("\n")[:2]
            info = info.replace(base_path, ".")
            info_lines = [x.strip(",") for x in info.strip().split(" ")]
            if "site-packages/" in info_lines[1]:
                info_lines[1] = '"' + info_lines[1].split("site-packages/")[1]
            tb[i] = (
                f"  {dark_white}File {bright_magenta}{info_lines[1]}"
                f"{dark_white}, line {bright_blue}{info_lines[3]}"
                f"{dark_white}, in {bright_cyan}{info_lines[5]}{BASE}m"
            )
            if code:
                tb[i] += f"\n{code}"

        msg = str(exc)
        if isinstance(exc, VyperException):
            # apply syntax highlight and remove traceback on vyper exceptions
            msg = self.highlight(msg)
            if not CONFIG.argv["tb"]:
                tb.clear()

        from brownie.exceptions import CompilerError

        if isinstance(exc, CompilerError):
            # apply syntax highlighting on solc exceptions
            if exc.compiler == "solc":
                msg = self.highlight(msg, SolidityLexer())
            else:
                msg = self.highlight(msg)
            if not CONFIG.argv["tb"]:
                tb.clear()

        tb.append(f"{bright_red}{type(exc).__name__}{BASE}m: {msg}")
        return "\n".join(tb)

    def format_syntaxerror(self, exc: SyntaxError) -> str:
        text = cast(str, exc.text)
        offset = cast(int, exc.offset) + len(text.lstrip()) - len(text) + 3
        exc.filename = cast(str, exc.filename).replace(base_path, ".")
        return (
            f'  {dark_white}File "{bright_magenta}{exc.filename}'
            f'{dark_white}", line {bright_blue}{exc.lineno}'
            f"{dark_white},\n{BASE}m    {text.strip()}\n"
            f"{' '*offset}^\n{bright_red}SyntaxError{BASE}m: {exc.msg}"
        )

    def highlight(self, text, lexer=PythonLexer()):
        """
        Apply syntax highlighting to a string.
        """
        return pygments.highlight(text, lexer, formatter)


def notify(type_: NotifyType, msg):
    """Prepends a message with a colored tag and outputs it to the console."""
    color = Color()
    print(f"{color(NOTIFY_COLORS[type_])}{type_}{BASE}m: {msg}")



#!/usr/bin/python3
import os
import pathlib
import shutil
import sys
import warnings
from typing import Any, DefaultDict, Final, Literal, NewType, final

import yaml
from dotenv import dotenv_values, load_dotenv
from hypothesis import Phase
from hypothesis import settings as hp_settings
from hypothesis.database import DirectoryBasedExampleDatabase
from mypy_extensions import mypyc_attr

from brownie._c_constants import Path, deepcopy, defaultdict, regex_sub, ujson_loads
from brownie._expansion import expand_posix_vars
from brownie._singleton import _Singleton
from brownie.typing import EvmVersion

__version__: Final = "1.22.0"

BROWNIE_FOLDER: Final = Path(sys.modules["brownie"].__file__).parent  # type: ignore [arg-type]
DATA_FOLDER: Final = Path.home().joinpath(".brownie")

DATA_SUBFOLDERS: Final = "accounts", "packages"

EVM_EQUIVALENTS: Final[dict[EvmVersion, EvmVersion]] = {
    EvmVersion("atlantis"): EvmVersion("byzantium"),
    EvmVersion("agharta"): EvmVersion("petersburg"),
}

python_version: Final = (
    f"{sys.version_info.major}.{sys.version_info.minor}"
    f".{sys.version_info.micro} {sys.version_info.releaselevel}"
)
REQUEST_HEADERS: Final = {"User-Agent": f"Brownie/{__version__} (Python/{python_version})"}


NetworkType = Literal["live", "development", None]
NetworkConfig = NewType("NetworkConfig", dict[str, Any])
# TODO: Make this a typed dict


@mypyc_attr(native_class=False)
class ConfigContainer:
    def __init__(self) -> None:
        base_config = _load_config(BROWNIE_FOLDER.joinpath("data/default-config.yaml"))
        if Path.home().joinpath("brownie-config.yaml").exists():
            home_config = _load_config(Path.home().joinpath("brownie-config.yaml"))
            _recursive_update(base_config, home_config)

        networks: dict[str, dict] = {}
        self.networks: Final = networks

        network_config = _load_config(_get_data_folder().joinpath("network-config.yaml"))
        for value in network_config["development"]:
            key = value["id"]
            if key in networks:
                raise ValueError(f"Multiple networks using ID '{key}'")
            networks[key] = value
        for value in [x for i in network_config["live"] for x in i["networks"]]:
            key = value["id"]
            if key in networks:
                raise ValueError(f"Multiple networks using ID '{key}'")
            networks[key] = value

        # make sure chainids are always strings
        for settings in networks.values():
            if "chainid" in settings:
                settings["chainid"] = str(settings["chainid"])

        self.argv: Final[DefaultDict[str, Any]] = defaultdict(_None_factory)
        self.settings: Final["ConfigDict"] = _Singleton("settings", (ConfigDict,), {})(base_config)
        self._active_network: NetworkConfig | None = None

        self.settings._lock()
        _modify_hypothesis_settings(self.settings["hypothesis"], "brownie-base", "default")

    def set_active_network(self, id_: str | None = None) -> NetworkConfig:
        """Modifies the 'active_network' configuration settings"""
        if id_ is None:
            id_ = self.settings["networks"]["default"]

        network = NetworkConfig(deepcopy(self.networks[id_]))
        key = "development" if "cmd" in network else "live"
        network["settings"] = self.settings["networks"][key].copy()

        if (
            key == "development"
            and isinstance(cmd_settings := network["cmd_settings"], dict)
            and "fork" in cmd_settings
        ):

            fork = cmd_settings["fork"]
            if fork in self.networks:
                fork_settings: dict = self.networks[fork]
                cmd_settings["fork"] = fork_settings["host"]
                network["chainid"] = fork_settings["chainid"]
                if "chain_id" not in cmd_settings:
                    cmd_settings["chain_id"] = int(fork_settings["chainid"])
                if "explorer" in fork_settings:
                    network["explorer"] = fork_settings["explorer"]

            cmd_settings["fork"] = os.path.expandvars(cmd_settings["fork"])

        self._active_network = network
        return network

    def clear_active(self) -> None:
        self._active_network = None

    @property
    def active_network(self) -> NetworkConfig:
        if self._active_network is None:
            raise ConnectionError("No active network")
        return self._active_network

    @property
    def network_type(self) -> NetworkType:
        if self._active_network is None:
            return None
        return "development" if "cmd" in self._active_network else "live"

    @property
    def mode(self) -> str | None:
        return self.argv["cli"]


@final
class Config(ConfigContainer, metaclass=_Singleton): ...


@final
class ConfigDict(dict[str, Any]):
    """Dict subclass that prevents adding new keys when locked"""

    def __init__(self, values: dict[str, Any] = {}) -> None:
        self._locked = False
        super().__init__()
        self.update(values)

    def __setitem__(self, key: str, value: Any) -> None:
        if self._locked and key not in self:
            raise KeyError(f"{key} is not a known config setting")
        if type(value) is dict:
            value = ConfigDict(value)
        super().__setitem__(key, value)

    def update(self, arg: dict[str, Any]) -> None:  # type: ignore [override]
        for k, v in arg.items():
            self.__setitem__(k, v)

    def _lock(self) -> None:
        """Locks the dict so that new keys cannot be added"""
        for obj in self.values():
            if type(obj) is ConfigDict:
                obj._lock()
        self._locked = True

    def _unlock(self) -> None:
        """Unlocks the dict so that new keys can be added"""
        for obj in self.values():
            if type(obj) is ConfigDict:
                obj._unlock()
        self._locked = False

    def _copy(self) -> dict[str, Any]:
        config_copy = {}
        for key, value in self.items():
            if isinstance(value, ConfigDict):
                value = value._copy()
            config_copy[key] = value
        return config_copy


def _get_project_config_path(project_path: pathlib.Path) -> pathlib.Path | None:
    if project_path.is_dir():
        path = project_path.joinpath("brownie-config")
    else:
        path = project_path
    suffix = next((i for i in (".yml", ".yaml", ".json") if path.with_suffix(i).exists()), None)
    return None if suffix is None else path.with_suffix(suffix)


def _load_config(project_path: pathlib.Path) -> dict:
    """Loads configuration data from a file, returns as a dict"""
    path = _get_project_config_path(project_path)
    if path is None:
        return {}

    with path.open() as fp:
        if path.suffix in (".yaml", ".yml"):
            return yaml.safe_load(fp) or {}
        raw_json = fp.read()
    valid_json = regex_sub(r'\/\/[^"]*?(?=\n|$)', "", raw_json)
    return ujson_loads(valid_json)


def _load_project_config(project_path: pathlib.Path) -> None:
    """Loads configuration settings from a project's brownie-config.yaml"""
    config_path = project_path.joinpath("brownie-config")
    config_data = _load_config(config_path)
    config_vars = _load_project_envvars(project_path)

    if "dotenv" in config_data:
        if not isinstance(config_data["dotenv"], str):
            raise ValueError(f'Invalid value passed to dotenv: {config_data["dotenv"]}')
        env_path = project_path.joinpath(config_data["dotenv"])
        if not env_path.is_file():
            raise ValueError(f"Dotenv specified in config but not found at path: {env_path}")
        config_vars.update(dotenv_values(dotenv_path=env_path))
        load_dotenv(dotenv_path=env_path)
        config_data = expand_posix_vars(config_data, config_vars)

    if not config_data:
        return

    if "network" in config_data:
        warnings.warn(
            "The `network` field in `brownie-config.yaml` has been deprecated. "
            "Network settings are now handled via `brownie networks` in the CLI. "
            f"Remove `network` from {config_path} to silence this warning.",
            DeprecationWarning,
        )
        del config_data["network"]

    # Update the network config cmd_settings with project specific cmd_settings
    if "networks" in config_data and isinstance(config_data["networks"], dict):
        for network, values in config_data["networks"].items():
            if (
                network != "default"
                and network in CONFIG.networks.keys()
                and "cmd_settings" in values
                and isinstance(values["cmd_settings"], dict)
            ):
                if "cmd_settings" in CONFIG.networks[network]:
                    _recursive_update(
                        CONFIG.networks[network]["cmd_settings"], values["cmd_settings"]
                    )
                else:
                    CONFIG.networks[network]["cmd_settings"] = values["cmd_settings"]

    settings = CONFIG.settings
    settings._unlock()
    _recursive_update(settings, config_data)
    _recursive_update(settings, expand_posix_vars(settings, config_vars))

    settings._lock()
    if "hypothesis" in config_data:
        _modify_hypothesis_settings(config_data["hypothesis"], "brownie", "brownie-base")


def _load_project_compiler_config(project_path: pathlib.Path | None) -> dict:
    if not project_path:
        return CONFIG.settings["compiler"]

    compiler_data = CONFIG.settings["compiler"]._copy()
    project_data = _load_config(project_path.joinpath("brownie-config")).get("compiler", {})
    _recursive_update(compiler_data, project_data)

    return compiler_data


def _load_project_envvars(project_path: pathlib.Path) -> dict:
    config_vars = dict(os.environ)
    settings = CONFIG.settings
    if settings.get("dotenv"):
        dotenv_path = settings["dotenv"]
        if not isinstance(dotenv_path, str):
            raise ValueError(f"Invalid value passed to dotenv: {dotenv_path}")
        env_path = project_path.joinpath(dotenv_path)
        if not env_path.is_file():
            raise ValueError(f"Dotenv specified in config but not found at path: {env_path}")
        config_vars.update(dotenv_values(dotenv_path=env_path))  # type: ignore [arg-type]
    return config_vars


def _load_project_structure_config(project_path):
    structure = CONFIG.settings["project_structure"]._copy()

    path = _get_project_config_path(project_path)
    if path is None:
        return structure

    data = _load_config(project_path).get("project_structure", {})
    structure.update(data)
    return structure


def _load_project_dependencies(project_path: pathlib.Path) -> list[str]:
    data = _load_config(project_path.joinpath("brownie-config"))
    dependencies = data.get("dependencies", []) or []
    if isinstance(dependencies, str):
        dependencies = [dependencies]
    return dependencies


def _modify_hypothesis_settings(settings, name, parent=None):
    settings = settings.copy()
    if parent is None:
        parent = hp_settings._current_profile  # type: ignore [attr-defined]

    if "phases" in settings:
        try:
            settings["phases"] = [getattr(Phase, k) for k, v in settings["phases"].items() if v]
        except AttributeError as exc:
            raise ValueError(f"'{exc.args[0]}' is not a valid hypothesis phase setting")

    hp_settings.register_profile(
        name,
        parent=hp_settings.get_profile(parent),
        database=DirectoryBasedExampleDatabase(str(_get_data_folder().joinpath("hypothesis"))),
        **settings,
    )
    hp_settings.load_profile(name)


def _recursive_update(original: dict, new: dict) -> None:
    """Recursively merges a new dict into the original dict"""
    if not original:
        original = {}
    for k in new:
        if k in original and isinstance(new[k], dict):
            _recursive_update(original[k], new[k])
        else:
            original[k] = new[k]


def _update_argv_from_docopt(args: dict[str, Any]) -> None:
    CONFIG.argv.update({k.lstrip("-"): v for k, v in args.items()})


def _get_data_folder() -> pathlib.Path:
    return DATA_FOLDER


def _make_data_folders(data_folder: pathlib.Path) -> None:
    # create data folder structure
    data_folder.mkdir(exist_ok=True)
    for folder in DATA_SUBFOLDERS:
        data_folder.joinpath(folder).mkdir(exist_ok=True)

    if not data_folder.joinpath("network-config.yaml").exists():
        shutil.copyfile(
            BROWNIE_FOLDER.joinpath("data/network-config.yaml"),
            data_folder.joinpath("network-config.yaml"),
        )

    if not data_folder.joinpath("providers-config.yaml").exists():
        shutil.copyfile(
            BROWNIE_FOLDER.joinpath("data/providers-config.yaml"),
            data_folder.joinpath("providers-config.yaml"),
        )


def _None_factory() -> None:
    return None


warnings.filterwarnings("once", category=DeprecationWarning, module="brownie")

# create data folders
_make_data_folders(DATA_FOLDER)

CONFIG: Final = Config()



from typing import TYPE_CHECKING, Any, Literal, NewType, TypedDict, TypeVar, final

from eth_event.main import EventData
from eth_typing import ABIElement, ChecksumAddress, HexStr
from typing_extensions import NotRequired

if TYPE_CHECKING:
    from brownie.network.account import Accounts
    from brownie.network.transaction import TransactionReceipt

# NETWORK
# Account
AccountsType = TypeVar("AccountsType", bound="Accounts")


# Contract
ContractName = NewType("ContractName", str)
FunctionName = NewType("FunctionName", str)
Selector = NewType("Selector", HexStr)


# Event
@final
class FormattedEvent(TypedDict):
    name: str | Literal["(anonymous)", "(unknown)"]
    data: list[EventData]
    decoded: bool
    address: ChecksumAddress


# Transactions
TransactionReceiptType = TypeVar("TransactionReceiptType", bound="TransactionReceipt")


# PROJECT
Start = int
Stop = int
Offset = tuple[Start, Stop]


# Build
class BytecodeJson(TypedDict):
    object: HexStr
    linkReferences: NotRequired[dict[str, dict]]


@final
class DeployedBytecodeJson(BytecodeJson):
    sourceMap: str
    opcodes: str


class _BuildJsonBase(TypedDict):
    contractName: ContractName
    abi: list[ABIElement]
    sha1: HexStr
    dependencies: NotRequired[list[ContractName]]


@final
class InterfaceBuildJson(_BuildJsonBase):
    type: Literal["interface"]
    source: str | None
    offset: Offset | None
    ast: NotRequired[list[dict]]


IntegerString = str
"""An integer cast as a string, as in: ``str(123)``"""

Statements = dict[str, dict["Count", Offset]]
StatementMap = dict[IntegerString, Statements]

Branches = dict[str, dict["Count", tuple[int, int, bool]]]
BranchMap = dict[IntegerString, Branches]


class CoverageMap(TypedDict):
    statements: StatementMap
    branches: BranchMap


class _ContractBuildJson(_BuildJsonBase):
    source: str
    sourcePath: str
    natspec: NotRequired[dict[str, Any]]
    allSourcePaths: dict[str, Any]
    offset: Offset
    bytecode: HexStr
    bytecodeSha1: HexStr
    deployedBytecode: HexStr
    coverageMap: CoverageMap
    pcMap: dict[int, "ProgramCounter"]
    compiler: NotRequired["CompilerConfig"]
    ast: NotRequired[list]


@final
class SolidityBuildJson(_ContractBuildJson):
    type: str
    language: Literal["Solidity"]
    opcodes: NotRequired[list[str]]
    sourceMap: NotRequired[dict]  # TODO: define typed dict
    deployedSourceMap: NotRequired[dict]  # TODO: define typed dict


@final
class VyperBuildJson(_ContractBuildJson):
    type: Literal["contract"]
    language: Literal["Vyper"]


ContractBuildJson = SolidityBuildJson | VyperBuildJson
BuildJson = ContractBuildJson | InterfaceBuildJson


# Compiler
Language = Literal["Solidity", "Vyper"]
EvmVersion = NewType("EvmVersion", str)
Source = tuple[Start, Stop, ContractName, str]


class ContractSource(TypedDict):
    content: str


class InterfaceSource(TypedDict):
    abi: list[ABIElement]


SourcesDict = dict[str, ContractSource | InterfaceSource]
InterfaceSources = dict[str, InterfaceSource]


@final
class OptimizerSettings(TypedDict):
    enabled: bool
    runs: int


@final
class SolcConfig(TypedDict):
    version: NotRequired[str | None]
    evm_version: NotRequired[EvmVersion]
    optimize: NotRequired[bool]
    runs: NotRequired[int]
    remappings: NotRequired[list[str] | None]
    optimizer: NotRequired[OptimizerSettings]
    viaIR: NotRequired[bool]


@final
class VyperConfig(TypedDict):
    version: NotRequired[str | None]
    evm_version: NotRequired[EvmVersion]


@final
class CompilerConfig(TypedDict):
    evm_version: EvmVersion | None
    solc: NotRequired[SolcConfig]
    vyper: VyperConfig
    version: NotRequired[str]
    optimizer: NotRequired[OptimizerSettings]


OutputSelection = dict[str, dict[str, list[str]]]


class _CompilerSettings(TypedDict):
    outputSelection: OutputSelection


@final
class SettingsSolc(_CompilerSettings):
    evmVersion: NotRequired[EvmVersion | None]
    remappings: list[str]
    optimizer: NotRequired[OptimizerSettings]
    viaIR: NotRequired[bool]


@final
class SettingsVyper(_CompilerSettings):
    evmVersion: NotRequired[EvmVersion]


class _InputJsonBase(TypedDict):
    sources: SourcesDict  # NOTE: should this be contract sources?
    interfaces: InterfaceSources

    # if I add a stub like this does it type check properly for members and fallbacks?
    def __getitem__(self, name: ContractName) -> dict[str, Any]: ...  # type: ignore [misc]


@final
class InputJsonSolc(_InputJsonBase, total=False):
    language: Literal["Solidity", None]
    settings: SettingsSolc


@final
class InputJsonVyper(_InputJsonBase, total=False):
    language: Literal["Vyper"]
    settings: SettingsVyper


InputJson = InputJsonSolc | InputJsonVyper


Count = int


class ProgramCounter(TypedDict):
    count: Count
    op: str
    fn: NotRequired[str | None]
    path: NotRequired[str]
    value: NotRequired[str]
    pc: NotRequired[int]
    branch: NotRequired[Count]
    jump: NotRequired[str]
    dev: NotRequired[str]
    offset: NotRequired[Offset]
    optimizer_revert: NotRequired[Literal[True]]
    first_revert: NotRequired[Literal[True]]
    jump_revert: NotRequired[Literal[True]]
    statement: NotRequired[Count]


PcList = list[ProgramCounter]
PCMap = NewType("PCMap", dict[int, ProgramCounter])


class VyperAstNode(TypedDict):
    """A dictionary representing on object on the AST."""

    name: str
    module: str
    type: str
    ast_type: str
    src: str
    op: "VyperAstNode"
    value: dict
    test: dict


VyperAstJson = list[VyperAstNode]



"""
This helper file exists so we can use `__file__`, which does not exist when a file is compiled with mypyc.
"""

from brownie._c_constants import Path, import_module
from brownie.network.middlewares import BrownieMiddlewareABC


def load_middlewares() -> list[type[BrownieMiddlewareABC]]:
    """
    Load middleware classes from all modules within `brownie/networks/middlewares/`.

    To be included the module name must not begin with `_` and the middleware
    must subclass :class:`~BrownieMiddlewareABC`.
    """
    middlewares = []
    for path in Path(__file__).parent.glob("[!_]*.py"):
        module = import_module(f"{__package__}.{path.stem}")
        middlewares.extend(
            obj
            for obj in module.__dict__.values()
            if isinstance(obj, type)
            and obj.__module__ == module.__name__
            and BrownieMiddlewareABC in obj.mro()
        )
    return middlewares



#!/usr/bin/python3

import sqlite3
import threading
from pathlib import Path
from typing import Any, Final, final

from brownie._c_constants import ujson_dumps, ujson_loads


@final
class Cursor:
    __slots__ = "_lock", "_db", "_cur", "_execute", "_fetchone", "_fetchall"

    def __init__(self, filename: Path) -> None:
        self._lock: Final = threading.Lock()
        self.connect(filename)

    def connect(self, filename: Path) -> None:
        self._db = sqlite3.connect(str(filename), isolation_level=None, check_same_thread=False)
        self._cur = self._db.cursor()
        self._execute = self._cur.execute
        self._fetchone = self._cur.fetchone
        self._fetchall = self._cur.fetchall

    def insert(self, table: str, *values: Any) -> None:
        encoded = [ujson_dumps(i) if isinstance(i, (dict, list)) else i for i in values]
        with self._lock:
            self._execute(
                f"INSERT OR REPLACE INTO {table} VALUES ({','.join('?'*len(encoded))})", encoded
            )

    def execute(self, cmd: str, *args: Any) -> None:
        with self._lock:
            self._execute(cmd, *args)

    def fetchone(self, cmd: str, *args: Any) -> tuple[Any, ...] | None:
        with self._lock:
            self._execute(cmd, *args)
            if result := self._fetchone():
                return tuple(ujson_loads(i) if str(i).startswith(("[", "{")) else i for i in result)
            return None

    def fetchall(self, cmd: str, *args: Any) -> Any:
        with self._lock:
            self._execute(cmd, *args)
            return self._fetchall()

    def close(self) -> None:
        self._cur.close()
        self._db.close()



from collections import defaultdict
from dataclasses import dataclass
from threading import Lock, get_ident
from types import FunctionType, TracebackType
from typing import Any

from lazy_object_proxy import Proxy
from wrapt import ObjectProxy

from brownie._c_constants import ujson_loads
from brownie._config import BROWNIE_FOLDER, CONFIG
from brownie.exceptions import ContractNotFound
from brownie.network import accounts, web3
from brownie.network.contract import Contract, ContractCall
from brownie.project import compile_source
from brownie.utils import color

DATA_DIR = BROWNIE_FOLDER.joinpath("data")
MULTICALL2_ABI = ujson_loads(DATA_DIR.joinpath("interfaces", "Multicall2.json").read_text())
MULTICALL2_SOURCE = DATA_DIR.joinpath("contracts", "Multicall2.sol").read_text()


@dataclass
class Call:

    calldata: tuple[str, bytes]
    decoder: FunctionType
    readable: str


class Result(ObjectProxy):
    """A proxy object to be updated with the result of a multicall."""

    def __repr__(self) -> str:
        return repr(self.__wrapped__)


class LazyResult(Proxy):
    """A proxy object to be updated with the result of a multicall."""

    def __repr__(self) -> str:
        return repr(self.__wrapped__)


class Multicall:
    """Context manager for batching multiple calls to constant contract functions."""

    _lock = Lock()

    def __init__(self) -> None:
        self.address = None
        self.default_verbose = False
        self._block_number = defaultdict(lambda: None)
        self._verbose = defaultdict(lambda: None)
        self._contract = None
        self._pending_calls: dict[int, list[Call]] = defaultdict(list)

        setattr(ContractCall, "__original_call_code", ContractCall.__call__.__code__)
        setattr(ContractCall, "__proxy_call_code", self._proxy_call.__code__)
        setattr(ContractCall, "__multicall", defaultdict(lambda: None))
        ContractCall.__call__.__code__ = self._proxy_call.__code__

    @property
    def block_number(self) -> int:
        return self._block_number[get_ident()]

    def __call__(
        self,
        address: str | None = None,
        block_identifier: str | bytes | int | None = None,
        verbose: bool | None = None,
    ) -> "Multicall":
        self.address = address
        self._block_number[get_ident()] = block_identifier
        self._verbose[get_ident()] = verbose if verbose is not None else self.default_verbose
        return self

    def _flush(self, future_result: Result = None) -> Any:
        pending_calls = self._pending_calls[get_ident()]
        self._pending_calls[get_ident()] = []

        if not pending_calls:
            # either all calls have already been made
            # or this result has already been retrieved
            return future_result
        with self._lock:
            if self._verbose.get(get_ident(), self.default_verbose):
                message = (
                    "Multicall:"
                    f"\n  Thread ID: {get_ident()}"
                    f"\n  Block number: {self._block_number[get_ident()]}"
                    f"\n  Calls: {len(pending_calls)}"
                )
                for c, item in enumerate(pending_calls, start=1):
                    u = "\u2514" if c == len(pending_calls) else "\u251c"
                    message = f"{message}\n    {u}\u2500{item.readable}"
                print(color.highlight(f"{message}\n"))

            ContractCall.__call__.__code__ = getattr(ContractCall, "__original_call_code")
            try:
                results = self._contract.tryAggregate(
                    False,
                    [_call.calldata for _call in pending_calls],
                    block_identifier=self._block_number[get_ident()],
                )
            finally:
                ContractCall.__call__.__code__ = getattr(ContractCall, "__proxy_call_code")

        for _call, result in zip(pending_calls, results):
            _call.__wrapped__ = _call.decoder(result[1]) if result[0] else None

        return future_result

    def flush(self) -> Any:
        """Flush the pending queue of calls, retrieving all the results."""
        return self._flush()

    def _call_contract(self, call: ContractCall, *args: Any, **kwargs: Any) -> Proxy:
        """Add a call to the buffer of calls to be made"""
        calldata = (call._address, call.encode_input(*args, **kwargs))
        readable = f"{call._name}({', '.join(str(i) for i in args)})"
        call_obj = Call(calldata, call.decode_output, readable)
        # future result
        result = Result(call_obj)
        self._pending_calls[get_ident()].append(result)

        return LazyResult(lambda: self._flush(result))

    @staticmethod
    def _proxy_call(*args: Any, **kwargs: Any) -> Any:
        """Proxy code which substitutes `ContractCall.__call__"""
        if self := getattr(ContractCall, "__multicall", {}).get(get_ident()):
            return self._call_contract(*args, **kwargs)

        # standard call we let pass through
        ContractCall.__call__.__code__ = getattr(ContractCall, "__original_call_code")
        try:
            result = ContractCall.__call__(*args, **kwargs)
        finally:
            ContractCall.__call__.__code__ = getattr(ContractCall, "__proxy_call_code")
        return result

    def __enter__(self) -> "Multicall":
        """Enter the Context Manager and substitute `ContractCall.__call__`"""
        # we set the code objects on ContractCall class so we can grab them later

        active_network = CONFIG.active_network

        if "multicall2" in active_network:
            self.address = active_network["multicall2"]
        elif "cmd" in active_network:
            deployment = self.deploy({"from": accounts[0]})
            self.address = deployment.address
            self._block_number[get_ident()] = deployment.tx.block_number

        self._block_number[get_ident()] = (
            self._block_number[get_ident()] or web3.eth.get_block_number()
        )

        if self.address is None:
            raise ContractNotFound(
                "Must set Multicall address via `brownie.multicall(address=...)`"
            )
        elif not web3.eth.get_code(self.address, block_identifier=self.block_number):
            raise ContractNotFound(
                f"Multicall at address {self.address} does not exist at block {self.block_number}"
            )

        self._contract = Contract.from_abi("Multicall", self.address, MULTICALL2_ABI)
        getattr(ContractCall, "__multicall")[get_ident()] = self

    def __exit__(self, exc_type: Exception, exc_val: Any, exc_tb: TracebackType) -> None:
        """Exit the Context Manager and reattach original `ContractCall.__call__` code"""
        self.flush()
        getattr(ContractCall, "__multicall")[get_ident()] = None

        self._block_number.pop(get_ident(), None)
        self._verbose.pop(get_ident(), None)

    @staticmethod
    def deploy(tx_params: dict) -> Contract:
        """Deploy an instance of the `Multicall2` contract.

        Args:
            tx_params: parameters passed to the `deploy` method of the `Multicall2` contract
                container.
        """
        project = compile_source(MULTICALL2_SOURCE)
        deployment = project.Multicall2.deploy(tx_params)
        CONFIG.active_network["multicall2"] = deployment.address
        return deployment



#!/usr/bin/python3
# mypy: disable-error-code="index"

from typing import Final, Literal, cast, final

from brownie.typing import (
    BuildJson,
    ContractBuildJson,
    ContractName,
    Count,
    InterfaceBuildJson,
    Language,
    Offset,
    PCMap,
    ProgramCounter,
)

from .sources import Sources, highlight_source

INTERFACE_KEYS: Final = "abi", "contractName", "sha1", "type"

DEPLOYMENT_KEYS: Final = (
    "abi",
    "ast",
    "bytecode",
    "compiler",
    "contractName",
    "deployedBytecode",
    "deployedSourceMap",
    "language",
    "natspec",
    "opcodes",
    "pcMap",
    "sourceMap",
    "type",
)

BUILD_KEYS: Final = (
    "allSourcePaths",
    "bytecodeSha1",
    "coverageMap",
    "dependencies",
    "offset",
    "sha1",
    "source",
    "sourcePath",
) + DEPLOYMENT_KEYS

_revert_map: Final[dict[int | str, tuple | Literal[False]]] = {}


@final
class Build:
    """Methods for accessing and manipulating a project's contract build data."""

    def __init__(self, sources: Sources) -> None:
        self._sources: Final = sources
        self._contracts: Final[dict[ContractName, ContractBuildJson]] = {}
        self._interfaces: Final[dict[ContractName, InterfaceBuildJson]] = {}

    def _add_contract(
        self,
        build_json: ContractBuildJson,
        alias: ContractName | None = None,
    ) -> None:
        contract_name = alias or build_json["contractName"]
        if contract_name in self._contracts and build_json["type"] == "interface":
            return
        if build_json["sourcePath"].startswith("interface"):
            # interfaces should generate artifact in /build/interfaces/ not /build/contracts/
            return
        self._contracts[contract_name] = build_json
        if "pcMap" not in build_json:
            # no pcMap means build artifact is for an interface
            return

        pc_map: dict[int | str, ProgramCounter] = build_json["pcMap"]  # type: ignore [assignment]
        if "0" in pc_map:
            build_json["pcMap"] = PCMap({Count(int(k)): pc_map[k] for k in pc_map})
        self._generate_revert_map(pc_map, build_json["allSourcePaths"], build_json["language"])

    def _add_interface(self, build_json: InterfaceBuildJson) -> None:
        contract_name = build_json["contractName"]
        self._interfaces[contract_name] = build_json

    def _generate_revert_map(
        self,
        pcMap: dict[int | str, ProgramCounter],
        source_map: dict[str, str],
        language: Language,
    ) -> None:
        # Adds a contract's dev revert strings to the revert map and it's pcMap
        marker = "//" if language == "Solidity" else "#"
        for pc, data in pcMap.items():
            op = data["op"]
            if op in ("REVERT", "INVALID") or "jump_revert" in data:
                path = data.get("path")
                if path is None:
                    continue

                path_str = source_map[path]

                if "dev" not in data:
                    if "fn" not in data or "first_revert" in data:
                        _revert_map[pc] = False
                        continue
                    try:
                        revert_str = self._sources.get(path_str)[data["offset"][1] :]
                        revert_str = revert_str[: revert_str.index("\n")]
                        revert_str = revert_str[revert_str.index(marker) + len(marker) :].strip()
                        if revert_str.startswith("dev:"):
                            data["dev"] = revert_str
                    except (KeyError, ValueError):
                        pass

                msg = "" if op == "REVERT" else "invalid opcode"
                revert = (
                    path_str,
                    data["offset"],
                    data.get("fn", "<None>"),
                    data.get("dev", msg),
                    self._sources,
                )

                # do not compare the final tuple item in case the same project was loaded twice
                if pc not in _revert_map or (
                    _revert_map[pc] and revert[:-1] == _revert_map[pc][:-1]
                ):
                    _revert_map[pc] = revert
                    continue
                _revert_map[pc] = False

    def _remove_contract(self, contract_name: ContractName) -> None:
        key = self._stem(contract_name)
        if key in self._contracts:
            del self._contracts[key]

    def _remove_interface(self, contract_name: ContractName) -> None:
        key = self._stem(contract_name)
        if key in self._interfaces:
            del self._interfaces[key]

    def get(self, contract_name: ContractName) -> BuildJson:
        """Returns build data for the given contract name."""
        key = self._stem(contract_name)
        if key in self._contracts:
            return self._contracts[key]
        return self._interfaces[key]

    def items(
        self,
        path: str | None = None,
    ) -> list[tuple[ContractName, BuildJson]]:
        """Provides an list of tuples as (key,value), similar to calling dict.items.
        If a path is given, only contracts derived from that source file are returned."""
        items = [*self._contracts.items(), *self._interfaces.items()]
        if path is None:
            return items
        return [(k, v) for k, v in items if v.get("sourcePath") == path]

    def contains(self, contract_name: ContractName) -> bool:
        """Checks if the contract name exists in the currently loaded build data."""
        stem = self._stem(contract_name)
        return stem in self._contracts or stem in self._interfaces

    def get_dependents(self, contract_name: ContractName) -> list[ContractName]:
        """Returns a list of contract names that inherit from or link to the given
        contract. Used by the compiler when determining which contracts to recompile
        based on a changed source file."""
        return [k for k, v in self._contracts.items() if contract_name in v.get("dependencies", [])]

    def _stem(self, contract_name: ContractName) -> ContractName:
        return contract_name.replace(".json", "")  # type: ignore [return-value]


def _get_dev_revert(pc: int) -> str | None:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the commented dev string (if any)
    if pc not in _revert_map:
        return None
    revert = _revert_map[pc]
    if revert is False:
        return None
    return revert[3]


def _get_error_source_from_pc(
    pc: int, pad: int = 3
) -> tuple[str | None, tuple[int, int] | None, str | None, str | None]:
    # Given the program counter from a stack trace that caused a transaction
    # to revert, returns the highlighted relevant source code and the method name.
    if pc not in _revert_map or _revert_map[pc] is False:
        return (None,) * 4
    revert = cast(tuple[str, Offset, str, str, Sources], _revert_map[pc])
    source = revert[4].get(revert[0])
    highlight, linenos = highlight_source(source, revert[1], pad=pad)
    return highlight, linenos, revert[0], revert[2]



from collections.abc import Callable, Sequence
from typing import Any, final

from web3 import Web3
from web3.types import RPCEndpoint

from brownie.network.middlewares import BrownieMiddlewareABC


@final
class Ganache7MiddleWare(BrownieMiddlewareABC):
    @classmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        return -100 if w3.client_version.lower().startswith("ganache/v7") else None

    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        result = make_request(method, params)

        # reformat failed eth_call / eth_sendTransaction output to mimic that of Ganache 6.x
        # yes, this is hacky and awful and in the future we should stop supporting
        # the older version of ganache. but doing so will cause unexpected issues
        # in projects that are still pinned to the old version, so for now we support
        # both and simply raise a warning of a pending deprecation.
        data: dict
        error: dict
        if (
            method in {"eth_sendTransaction", "eth_sendRawTransaction"}
            and "error" in result
            and "data" in (error := result["error"])
        ):
            data = error["data"]
            data["error"] = data.pop("message")
            data["program_counter"] = data.pop("programCounter")
            error["data"] = {data.pop("hash"): data}

        if method == "eth_call" and "error" in result:
            error = result["error"]
            if error.get("message", "").startswith("VM Exception"):
                # "VM Exception while processing transaction: {reason} {message}"
                msg: str = error["message"]
                msg = msg.split(": ", maxsplit=1)[-1]
                if msg.startswith("revert"):
                    data = {"error": "revert", "reason": error["data"]}
                else:
                    data = {"error": msg, "reason": None}
                error["data"] = {"0x": data}

        return result



#!/usr/bin/python3

import sys
from pathlib import Path

import pytest

from brownie import project
from brownie._config import CONFIG, _modify_hypothesis_settings
from brownie.test.fixtures import PytestBrownieFixtures
from brownie.test.managers import PytestBrownieMaster, PytestBrownieRunner, PytestBrownieXdistRunner
from brownie.utils import color


def _get_project_path() -> Path | None:
    key = next((i for i in sys.argv if i.startswith("--brownie-project")), "")
    if key == "--brownie-project":
        idx = sys.argv.index(key)
        project_path = Path(sys.argv[idx + 1]).absolute()
    elif key.startswith("--brownie-project="):
        project_path = Path(key[18:]).absolute()
    else:
        return project.check_for_project(".")

    if project_path != project.check_for_project(project_path):
        raise pytest.UsageError(f"Unable to load project at '{sys.argv[idx + 1]}'")
    return project_path


# set commandline options
def pytest_addoption(parser):
    parser.addoption(
        "--brownie-project", nargs=1, default=".", help="Load a brownie project at the given path"
    )

    if _get_project_path():
        parser.addoption(
            "--coverage", "-C", action="store_true", help="Evaluate contract test coverage"
        )
        parser.addoption(
            "--gas", "-G", action="store_true", help="Display gas profile for function calls"
        )
        parser.addoption(
            "--update", "-U", action="store_true", help="Only run tests where changes have occurred"
        )
        parser.addoption(
            "--revert-tb", "-R", action="store_true", help="Show detailed traceback on tx reverts"
        )
        parser.addoption(
            "--interactive",
            "-I",
            action="store_true",
            help="Open an interactive console each time a test fails",
        )
        parser.addoption(
            "--stateful",
            choices=["true", "false"],
            default=None,
            help="Only run or skip stateful tests (default: run all tests)",
        )
        parser.addoption(
            "--failfast", action="store_true", help="Fail hypothesis tests quickly (no shrinking)"
        )
        parser.addoption(
            "--network",
            "-N",
            default=False,
            nargs=1,
            help=f"Use a specific network (default {CONFIG.settings['networks']['default']})",
        )
        parser.addoption(
            "--showinternal",
            action="store_true",
            help="Include Brownie internal frames in tracebacks",
        )


def pytest_load_initial_conftests(early_config):
    capsys = early_config.pluginmanager.get_plugin("capturemanager")

    if project_path := _get_project_path():
        # suspend stdout capture to display compilation data
        capsys.suspend()
        try:
            active_project = project.load(project_path)

            active_project.load_config()
            active_project._add_to_main_namespace()
        except Exception as e:
            # prevent pytest INTERNALERROR traceback when project fails to compile
            print(f"{color.format_tb(e)}\n")
            raise pytest.UsageError("Unable to load project")
        finally:
            capsys.resume()


def pytest_configure(config):
    if not _get_project_path():
        return

    if not config.getoption("showinternal"):
        # do not include brownie internals in tracebacks
        base_path = Path(sys.modules["brownie"].__file__).parent.as_posix()
        for module in sys.modules.values():
            if getattr(module, "__file__", None) and module.__file__.startswith(base_path):
                module.__tracebackhide__ = True
                module.__hypothesistracebackhide__ = True

    # enable verbose output if stdout capture is disabled
    if config.getoption("capture") == "no":
        config.option.verbose = True

    # if verbose mode is enabled, also enable hypothesis verbose mode
    if config.option.verbose:
        _modify_hypothesis_settings({"verbosity": 2}, "brownie-verbose")

    if config.getoption("numprocesses"):
        if config.getoption("interactive"):
            raise ValueError("Cannot use --interactive mode with xdist")
        Plugin = PytestBrownieMaster
    elif hasattr(config, "workerinput"):
        Plugin = PytestBrownieXdistRunner
    else:
        Plugin = PytestBrownieRunner

    if config.getoption("interactive"):
        config.option.failfast = True

    if config.getoption("failfast"):
        _modify_hypothesis_settings(
            {"phases": {"explicit": True, "generate": True, "target": True}}, "brownie-failfast"
        )

    active_project = project.get_loaded_projects()[0]
    session = Plugin(config, active_project)
    config.pluginmanager.register(session, "brownie-core")

    if not config.getoption("numprocesses"):
        fixtures = PytestBrownieFixtures(config, active_project)
        config.pluginmanager.register(fixtures, "brownie-fixtures")



#!/usr/bin/python3

from typing import Final

from . import hex
from ._color import Color, notify

color: Final = Color()
bytes_to_hexstring: Final = hex.bytes_to_hexstring
hexbytes_to_hexstring: Final = hex.hexbytes_to_hexstring

__all__ = ["Color", "color", "notify", "bytes_to_hexstring", "hexbytes_to_hexstring"]



import threading
import time
from collections import OrderedDict
from collections.abc import Callable, Sequence
from typing import Any, Final, final

import faster_hexbytes
from web3 import Web3
from web3.types import LogReceipt, RPCEndpoint

from brownie._c_constants import HexBytes, ujson_dumps
from brownie._config import CONFIG, _get_data_folder
from brownie.network.middlewares import BrownieMiddlewareABC
from brownie.utils.sql import Cursor

# calls to the following RPC endpoints are stored in a persistent cache
# if the returned data evaluates true when passed into the lambda
LONGTERM_CACHE: Final = {
    "eth_getCode": lambda w3, data: is_cacheable_bytecode(w3, data),
}


def _strip_push_data(bytecode: bytes) -> bytes:
    idx = 0
    while idx < len(bytecode):
        # if instruction is between PUSH1 and PUSH32
        if 0x60 <= bytecode[idx] <= 0x7F:
            offset = idx + 1
            length = bytecode[idx] - 0x5F
            bytecode = bytecode[:offset] + bytecode[offset + length :]
        idx += 1
    return bytecode


def is_cacheable_bytecode(web3: Web3, bytecode: faster_hexbytes.HexBytes) -> bool:
    """
    Check if bytecode can safely by cached.

    To safely cache bytecode we verify that the code cannot be removed via a
    SELFDESTRUCT operation, or a SELFDESTRUCT triggered via a DELEGATECALL.

    Arguments
    ---------
    web3 : Web3
        Web3 object connected to the same network that the bytecode exists on.
    bytecode : HexBytes
        Deployed bytecode to be analyzed.

    Returns
    -------
    bool
        Can this bytecode be cached?
    """
    if not bytecode:
        # do not cache empty code, something might be deployed there later!
        return False

    bytecode_bytes = bytes(HexBytes(bytecode))
    opcodes = _strip_push_data(bytecode_bytes)
    if 0xFF in opcodes:
        # cannot cache if the code contains a SELFDESTRUCT instruction
        return False
    for idx in [i for i in range(len(opcodes)) if opcodes[i] == 0xF4]:
        # cannot cache if the code performs a DELEGATECALL to a not-fixed address
        if idx < 2:
            return False
        if opcodes[idx - 2 : idx] != b"sZ":  # b"sZ" == bytes(HexBytes("0x735A"))
            # if the instruction not is immediately preceded by PUSH20 GAS
            # the target was not hardcoded and we cannot cache
            return False

    # check if the target code of each delegatecall is also cacheable
    # if yes then we can cache this contract as well
    push20_indexes = (
        i
        for i in range(len(bytecode_bytes) - 22)
        if bytecode_bytes[i] == 0x73 and bytecode_bytes[i + 22] == 0xF4
    )
    for address in (bytecode_bytes[i + 1 : i + 21] for i in push20_indexes):
        if not int(address.hex(), 16):
            # if the delegatecall targets 0x00 this is a factory pattern, we can ignore
            continue
        target_bytecode = web3.eth.get_code(address)
        if not is_cacheable_bytecode(web3, target_bytecode):
            return False

    return True


def _new_filter(w3: Web3) -> Any:
    # returns a filter if the client is connected and supports filtering
    try:
        block_filter = w3.eth.filter("latest")
        block_filter.get_new_entries()
        return block_filter
    except (AttributeError, ValueError):
        return None


@final
class RequestCachingMiddleware(BrownieMiddlewareABC):
    """
    Web3 middleware for request caching.
    """

    def __init__(self, w3: Web3) -> None:
        super().__init__(w3)

        self.table_key: Final = f"chain{CONFIG.active_network['chainid']}"
        self.cur: Final = Cursor(_get_data_folder().joinpath("cache.db"))
        self.cur.execute(f"CREATE TABLE IF NOT EXISTS {self.table_key} (method, params, result)")

        self.lock: Final = threading.Lock()
        self.event: Final = threading.Event()
        self.start_block_filter_loop()

    def start_block_filter_loop(self):
        self.event.clear()
        self.loop_thread = threading.Thread(target=self.loop_exception_handler, daemon=True)
        self.loop_thread.start()
        self.event.wait()

    @classmethod
    def get_layer(cls, w3: Web3, network_type: str) -> int | None:
        if CONFIG.settings["eager_caching"] is False:
            # do not cache when user doesn't want it
            return None
        if network_type != "live":
            # do not cache on development chains
            return None
        try:
            latest = w3.eth.get_block("latest")
        except Exception:
            return None
        if latest.timestamp - w3.eth.get_block(latest.number - 50).timestamp < 250:
            # do not cache on chains with an average block time of less than 5 seconds
            return None
        if _new_filter(w3) is None:
            # do not cache if we cannot create a filter for new blocks
            return None
        return 0

    @property
    def time_since(self) -> float:
        return time.time() - self.last_request

    def loop_exception_handler(self) -> None:
        try:
            self.block_filter_loop()
        except Exception:
            # catch unhandled exceptions to avoid random error messages in the console
            self.block_cache.clear()
            self.is_killed = True

    def block_filter_loop(self) -> None:
        # initialize required state variables within the loop to avoid recursion death
        latest = self.w3.eth.get_block("latest")
        self.last_block = latest.hash
        self.last_block_seen = latest.timestamp
        self.last_request = time.time()
        self.block_cache: OrderedDict = OrderedDict()
        self.block_filter = self.w3.eth.filter("latest")
        self.is_killed = False
        self.event.set()

        new_blocks: list[LogReceipt]
        while not self.is_killed:
            # if the last RPC request was > 60 seconds ago, reduce the rate of updates.
            # we eventually settle at one query per minute after 10 minutes of no requests.
            with self.lock:
                if self.time_since > 60:
                    self.block_cache.clear()
                    self.event.clear()
            if self.time_since > 60:
                self.event.wait(min(self.time_since / 10, 60))

            # query the filter for new blocks
            with self.lock:
                try:
                    new_blocks = self.block_filter.get_new_entries()
                except (AttributeError, ValueError):
                    # web3 has disconnected, or the filter has expired from inactivity
                    # some public nodes allow a filter initially, but block it several seconds later
                    block_filter = _new_filter(self.w3)
                    if block_filter is None:
                        return
                    self.block_filter = block_filter

                    # continue in try: except: block is not supported by mypyc
                    # as of jul 23 2025 so we use this workaround instead.
                    should_skip = True
                else:
                    should_skip = False
                    if new_blocks:
                        self.block_cache[new_blocks[-1]] = {}
                        self.last_block = new_blocks[-1]
                        self.last_block_seen = time.time()
                        if len(self.block_cache) > 5:
                            old_key = list(self.block_cache)[0]
                            del self.block_cache[old_key]

            # continue in try: except: block is not supported by mypyc
            # as of jul 23 2025 so we use this workaround instead.
            if should_skip:
                pass

            elif new_blocks and self.time_since < 15:
                # if this update found a new block and we've been querying
                # frequently, we can wait a few seconds before the next update
                time.sleep(5)
            elif time.time() - self.last_block_seen < 15:
                # if it's been less than 15 seconds since the last block, wait 2 seconds
                time.sleep(2)
            else:
                # if it's been more than 15 seconds, only wait 1 second
                time.sleep(1)

    def process_request(
        self,
        make_request: Callable,
        method: RPCEndpoint,
        params: Sequence[Any],
    ) -> dict[str, Any]:
        if method in (
            # caching any of these means we die of recursion death so let's not do that
            "eth_getFilterChanges",
            "eth_newBlockFilter",
            "eth_uninstallFilter",
            # used to check connectivity
            "web3_clientVersion",
            # caching these causes weirdness with transaction broadcasting and replacement
            "eth_sendTransaction",
            "eth_sendRawTransaction",
            "eth_sign",
            "eth_signTransaction",
            "eth_getTransactionByHash",
            "eth_getTransactionReceipt",
            "eth_chainId",
        ):
            return make_request(method, params)

        # try to return a cached value
        param_str = ujson_dumps(params, separators=(",", ""), default=str)

        # check if the value is available within the long-term cache
        if method in LONGTERM_CACHE:
            row = self.cur.fetchone(
                f"SELECT result FROM {self.table_key} WHERE method=? AND params=?",
                (method, param_str),
            )
            if row:
                data = row[0]
                if isinstance(data, bytes):
                    data = HexBytes(data)
                return {"id": "cache", "jsonrpc": "2.0", "result": data}

        if not self.loop_thread.is_alive():
            # restart the block filter loop if it has crashed (usually from a ConnectionError)
            self.start_block_filter_loop()

        with self.lock:
            self.last_request = time.time()
            self.event.set()
            try:
                return self.block_cache[self.last_block][method][param_str]
            except KeyError:
                pass

        # cached value is unavailable, make a request and cache the result
        with self.lock:
            response = make_request(method, params)
            self.block_cache.setdefault(self.last_block, {}).setdefault(method, {})
            self.block_cache[self.last_block][method][param_str] = response

        # check if the value can be added to long-term cache
        if "result" in response and method in LONGTERM_CACHE:
            result = response["result"]
            if LONGTERM_CACHE[method](self.w3, result):
                if isinstance(result, (dict, list, tuple)):
                    result = ujson_dumps(response, separators=(",", ""), default=str)
                self.cur.insert(self.table_key, method, param_str, result)

        return response

    def uninstall(self) -> None:
        self.is_killed = True
        self.block_cache.clear()
        if self.w3.isConnected():
            self.w3.eth.uninstall_filter(self.block_filter.filter_id)



#######################################################################
# Implements a topological sort algorithm.
#
# Copyright 2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Notes:
#  Based on http://code.activestate.com/recipes/578272-topological-sort
#   with these major changes:
#    Added unittests.
#    Deleted doctests (maybe not the best idea in the world, but it cleans
#     up the docstring).
#    Moved functools import to the top of the file.
#    Changed assert to a ValueError.
#    Changed iter[items|keys] to [items|keys], for python 3
#     compatibility. I don't think it matters for python 2 these are
#     now lists instead of iterables.
#    Copy the input so as to leave it unmodified.
#    Renamed function from toposort2 to toposort.
#    Handle empty input.
#    Switch tests to use set literals.
#
########################################################################

import functools
from collections.abc import Iterator
from typing import Final, final

__all__ = ["toposort", "toposort_flatten", "CircularDependencyError"]


_reduce: Final = functools.reduce


@final
class CircularDependencyError(ValueError):
    def __init__(self, data):
        # Sort the data just to make the output consistent, for use in
        #  error messages.  That's convenient for doctests.
        super().__init__(
            "Circular dependencies exist among these items: {{{}}}".format(
                ", ".join(f"{key!r}:{value!r}" for key, value in sorted(data.items()))
            )
        )
        self.data = data


def toposort(data: dict) -> Iterator[set]:
    """Dependencies are expressed as a dictionary whose keys are items
    and whose values are a set of dependent items. Output is a list of
    sets in topological order. The first set consists of items with no
    dependencies, each subsequent set consists of items that depend upon
    items in the preceding sets."""

    # Special case empty input.
    if len(data) == 0:
        return

    # Copy the input so as to leave it unmodified.
    data = data.copy()

    # Ignore self dependencies.
    for k, v in data.items():
        v.discard(k)
    # Find all items that don't depend on anything.
    extra_items_in_deps = _reduce(set.union, data.values()) - set(data.keys())
    # Add empty dependencies where needed.
    data.update({item: set() for item in extra_items_in_deps})
    while True:
        ordered = {item for item, dep in data.items() if len(dep) == 0}
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered) for item, dep in data.items() if item not in ordered}
    if len(data) != 0:
        raise CircularDependencyError(data)


def toposort_flatten(data: dict, sort: bool = True) -> list:
    """Returns a single list of dependencies. For any set returned by
    toposort(), those items are sorted and appended to the result (just to
    make the results deterministic)."""

    result = []
    if sort:
        for d in toposort(data):
            result.extend(sorted(d))
    else:
        for d in toposort(data):
            result.extend(d)
    return result



#!/usr/bin/python3

import datetime
import re
import sys
import warnings
from subprocess import DEVNULL, PIPE
from typing import Any

import psutil
from eth_utils.toolz import concat
from requests.exceptions import ConnectionError as RequestsConnectionError

from brownie._config import EVM_EQUIVALENTS
from brownie.convert import Wei
from brownie.exceptions import InvalidArgumentWarning, RPCRequestError
from brownie.network.web3 import web3
from brownie.utils import bytes_to_hexstring

CLI_FLAGS = {
    "7": {
        "port": "--server.port",
        "gas_limit": "--miner.blockGasLimit",
        "accounts": "--wallet.totalAccounts",
        "evm_version": "--hardfork",
        "fork": "--fork.url",
        "disable_cache": "--fork.disableCache",
        "mnemonic": "--wallet.mnemonic",
        "account_keys_path": "--wallet.accountKeysPath",
        "block_time": "--miner.blockTime",
        "default_balance": "--wallet.defaultBalance",
        "time": "--chain.time",
        "unlock": "--wallet.unlockedAccounts",
        "network_id": "--chain.networkId",
        "chain_id": "--chain.chainId",
        "unlimited_contract_size": "--chain.allowUnlimitedContractSize",
    },
    "<=6": {
        "port": "--port",
        "gas_limit": "--gasLimit",
        "accounts": "--accounts",
        "evm_version": "--hardfork",
        "fork": "--fork",
        "mnemonic": "--mnemonic",
        "account_keys_path": "--acctKeys",
        "block_time": "--blockTime",
        "default_balance": "--defaultBalanceEther",
        "time": "--time",
        "unlock": "--unlock",
        "network_id": "--networkId",
        "chain_id": "--chainId",
        "unlimited_contract_size": "--allowUnlimitedContractSize",
    },
}

EVM_VERSIONS = ["byzantium", "constantinople", "petersburg", "istanbul"]
EVM_DEFAULT = "istanbul"


def launch(cmd: str, **kwargs: Any) -> None:
    """Launches the RPC client.

    Args:
        cmd: command string to execute as subprocess"""
    if sys.platform == "win32" and not cmd.split(" ")[0].endswith(".cmd"):
        if " " in cmd:
            cmd = cmd.replace(" ", ".cmd ", 1)
        else:
            cmd += ".cmd"
    cmd_list = cmd.split(" ")
    ganache_version = get_ganache_version(cmd_list[0])

    if ganache_version <= 6:
        cli_flags = CLI_FLAGS["<=6"]
    else:
        cli_flags = CLI_FLAGS["7"]
        # this flag must be true so that reverting tx's return a
        # more verbose output similar to what ganache 6 produced
        cmd_list.extend(["--chain.vmErrorsOnRPCResponse", "true"])

    kwargs.setdefault("evm_version", EVM_DEFAULT)
    if kwargs["evm_version"] in EVM_EQUIVALENTS:
        kwargs["evm_version"] = EVM_EQUIVALENTS[kwargs["evm_version"]]
    kwargs = _validate_cmd_settings(kwargs)
    for key, value in [(k, v) for k, v in kwargs.items() if v]:
        if key == "unlock":
            if not isinstance(value, list):
                value = [value]
            for address in value:
                if isinstance(address, int):
                    address = bytes_to_hexstring(address.to_bytes(20, "big"))
                cmd_list.extend([cli_flags[key], address])
        else:
            try:
                if value is True:
                    cmd_list.append(cli_flags[key])
                elif value is not False:
                    cmd_list.extend([cli_flags[key], str(value)])
            except KeyError:
                warnings.warn(
                    f"Ignoring invalid commandline setting for ganache-cli: "
                    f'"{key}" with value "{value}".',
                    InvalidArgumentWarning,
                )
    print(f"\nLaunching '{' '.join(cmd_list)}'...")
    out = DEVNULL if sys.platform == "win32" else PIPE

    return psutil.Popen(cmd_list, stdin=DEVNULL, stdout=out, stderr=out)


def get_ganache_version(ganache_executable: str) -> int:
    ganache_version_proc = psutil.Popen([ganache_executable, "--version"], stdout=PIPE)
    ganache_version_stdout, _ = ganache_version_proc.communicate()
    if ganache_version_match := re.search(r"v([0-9]+)\.", ganache_version_stdout.decode()):
        return int(ganache_version_match[1])
    raise ValueError(f"could not read ganache version: {ganache_version_stdout}")


def on_connection() -> None:
    pass


def _request(method: str, args: list) -> int:
    try:
        response = web3.provider.make_request(method, args)
        if "result" in response:
            return response["result"]
    except (AttributeError, RequestsConnectionError):
        raise RPCRequestError("Web3 is not connected.")
    raise RPCRequestError(response["error"]["message"])


def sleep(seconds: int) -> int:
    return _request("evm_increaseTime", [seconds])


def mine(timestamp: int | None = None) -> None:
    params = [timestamp] if timestamp else []
    _request("evm_mine", params)
    if timestamp and web3.client_version.lower().startswith("ganache/v7"):
        # ganache v7 does not modify the internal time when mining new blocks
        # so we also set the time to maintain consistency with v6 behavior
        _request("evm_setTime", [(timestamp + 1) * 1000])


def snapshot() -> int:
    return _request("evm_snapshot", [])


def revert(snapshot_id: int) -> None:
    _request("evm_revert", [snapshot_id])


def unlock_account(address: str) -> None:
    if web3.client_version.lower().startswith("ganache/v7"):
        web3.provider.make_request("evm_addAccount", [address, ""])
        web3.provider.make_request(
            "personal_unlockAccount",
            [address, "", 9999999999],
        )
    else:
        web3.provider.make_request("evm_unlockUnknownAccount", [address])


def _validate_cmd_settings(cmd_settings: dict) -> dict:
    ganache_keys = set(concat(CLI_FLAGS.values()))

    CMD_TYPES = {
        "port": int,
        "gas_limit": int,
        "block_time": int,
        "time": datetime.datetime,
        "accounts": int,
        "evm_version": str,
        "mnemonic": str,
        "account_keys_path": str,
        "fork": str,
        "disable_cache": bool,
        "network_id": int,
        "chain_id": int,
    }
    for cmd, value in cmd_settings.items():
        if cmd in ganache_keys and cmd in CMD_TYPES and not isinstance(value, CMD_TYPES[cmd]):
            raise TypeError(
                f'Wrong type for cmd_settings "{cmd}": {value}. '
                f"Found {type(value).__name__}, but expected {CMD_TYPES[cmd].__name__}."
            )

    if "default_balance" in cmd_settings:
        try:
            cmd_settings["default_balance"] = int(cmd_settings["default_balance"])
        except ValueError:
            # convert any input to ether, then format it properly
            default_eth = Wei(cmd_settings["default_balance"]).to("ether")
            cmd_settings["default_balance"] = (
                default_eth.quantize(1) if default_eth > 1 else default_eth.normalize()
            )
    return cmd_settings



#!/usr/bin/python3

import inspect
import sys

from brownie import network, project
from brownie._c_constants import Path
from brownie._cli.console import Console
from brownie._config import CONFIG, _update_argv_from_docopt
from brownie.project.scripts import _get_path, run
from brownie.test.output import _build_gas_profile_output
from brownie.utils import color
from brownie.utils.docopt import docopt

__doc__ = f"""Usage: brownie run <filename> [<function>] [<arg>...] [options]

Arguments:
  <filename>              The name of the script to run
  [<function>]            The function to call (default is main)
  [<arg>]                 Extra argument to pass to the function

Options:
  --network [name]        Use a specific network (default {CONFIG.settings['networks']['default']})
  --silent                Suppress console output for transactions
  --interactive -I        Open an interactive console when the script completes or raises
  --raise -r              Raise exceptions occurred in the script to the caller
  --gas -g                Display gas profile for function calls
  --tb -t                 Show entire python traceback on exceptions
  --help -h               Display this message

Use run to execute scripts for contract deployment, to automate common
interactions, or for gas profiling."""


def main():
    args = docopt(__doc__)
    _update_argv_from_docopt(args)

    active_project = None
    if project.check_for_project():
        active_project = project.load()
        active_project.load_config()
        print(f"{active_project._name} is the active project.")

    network.connect(CONFIG.argv["network"])

    path, _ = _get_path(args["<filename>"])
    path_str = path.absolute().as_posix()

    try:
        return_value, frame = run(
            args["<filename>"],
            method_name=args["<function>"] or "main",
            args=args["<arg>"],
            _include_frame=True,
        )
        exit_code = 0
    except Exception as e:
        if args["--raise"]:
            raise e

        print(color.format_tb(e))

        frame = next(
            (i.frame for i in inspect.trace()[::-1] if Path(i.filename).as_posix() == path_str),
            None,
        )
        if frame is None:
            # exception was an internal brownie issue - do not open the console
            sys.exit(1)
        exit_code = 1
        return_value = None

    try:
        if args["--interactive"]:
            # filter internal objects from the namespace prior to opening the console
            globals_dict = {k: v for k, v in frame.f_globals.items() if not k.startswith("__")}
            extra_locals = {"_": return_value, **globals_dict, **frame.f_locals}
            shell = Console(active_project, extra_locals)
            shell.interact(banner="\nInteractive mode enabled. Use quit() to close.", exitmsg="")
    finally:
        # the console terminates from a SystemExit - make sure we still deliver the final gas report
        if CONFIG.argv["gas"]:
            print("\n======= Gas profile =======")
            for line in _build_gas_profile_output():
                print(line)

        sys.exit(exit_code)



#!/usr/bin/python3

from typing import Final

OUTCOMES: Final = (
    (".", "passed"),
    ("s", "skipped"),
    ("F", "failed"),
    ("E", "error"),
    ("x", "xfailed"),
    ("X", "xpassed"),
)


def convert_outcome(value: str) -> str:
    return next(next(x for x in i if x != value) for i in OUTCOMES if value in i)



#!/usr/bin/python3

import sys
import threading
import time
from collections.abc import Iterator
from getpass import getpass
from importlib.metadata import version
from pathlib import Path
from typing import Any, Optional, Union

import eth_account
import eth_keys
import rlp
from eip712.messages import EIP712Message
from eth_account._utils.signing import sign_message_hash
from eth_account.datastructures import SignedMessage
from eth_account.messages import _hash_eip191_message, defunct_hash_message
from eth_typing import BlockNumber, HexAddress
from faster_eth_utils import keccak
from faster_eth_utils.applicators import apply_formatters_to_dict
from web3 import HTTPProvider, IPCProvider
from web3.exceptions import InvalidTransaction, TransactionNotFound

from brownie._c_constants import HexBytes, deque, ujson_dump, ujson_load
from brownie._config import CONFIG, _get_data_folder
from brownie._singleton import _Singleton
from brownie.convert import EthAddress, Wei, to_address
from brownie.exceptions import (
    ContractNotFound,
    TransactionError,
    UnknownAccount,
    VirtualMachineError,
)
from brownie.utils import bytes_to_hexstring, color
from brownie.utils._color import bright_blue, bright_cyan

from .gas.bases import GasABC
from .rpc import Rpc
from .state import Chain, TxHistory, _revert_register
from .transaction import TransactionReceipt
from .web3 import _resolve_address, web3

ETH_ACCOUNT_LT_0_13_0 = tuple(map(int, version("eth_account").split("."))) < (
    0,
    13,
    0,
)

history = TxHistory()
rpc = Rpc()

eth_account.Account.enable_unaudited_hdwallet_features()
_marker = deque("-/|\\-/|\\")


class Accounts(metaclass=_Singleton):
    """
    List-like container that holds all available `Account` objects.

    Attributes
    ----------
    default : Account, optional
        Default account to broadcast transactions from.
    """

    def __init__(self) -> None:
        self.default = None
        self._accounts: list = []

        # prevent sensitive info from being stored in readline history
        self.add.__dict__["_private"] = True
        self.from_mnemonic.__dict__["_private"] = True

        _revert_register(self)
        self._reset()

    def _reset(self) -> None:
        self._accounts.clear()
        try:
            self._accounts = [Account(i) for i in web3.eth.accounts]
        except Exception:
            pass

        # Check if accounts were manually unlocked and add them
        try:
            unlocked_accounts = CONFIG.active_network["cmd_settings"]["unlock"]
            if not isinstance(unlocked_accounts, list):
                unlocked_accounts = [unlocked_accounts]
            for address in unlocked_accounts:
                if isinstance(address, int):
                    address = bytes_to_hexstring(address.to_bytes(20, "big"))
                account = Account(address)
                if account not in self._accounts:
                    self._accounts.append(account)
        except (ConnectionError, ValueError, KeyError):
            pass

        if self.default not in self._accounts:
            self.default = None

    def _revert(self, height: BlockNumber) -> None:
        # must exist for rpc registry callback
        pass

    def __contains__(self, address: HexAddress) -> bool:
        try:
            address = to_address(address)
            return address in self._accounts
        except ValueError:
            return False

    def __repr__(self) -> str:
        if CONFIG.argv["cli"] == "console":
            return str(self._accounts)
        return super().__repr__()

    def __iter__(self) -> Iterator:
        return iter(self._accounts)

    def __getitem__(self, key: int) -> "_PrivateKeyAccount":
        return self._accounts[key]

    def __delitem__(self, key: int) -> None:
        del self._accounts[key]

    def __len__(self) -> int:
        return len(self._accounts)

    def add(self, private_key: int | bytes | str = None) -> "LocalAccount":
        """
        Create a new ``LocalAccount`` instance and appends it to the container.

        When the no private key is given, a mnemonic is also generated and outputted.

        Arguments
        ---------
        private_key : int | bytes | str, optional
            Private key of the account. If none is given, one is randomly generated.

        Returns
        -------
        LocalAccount
        """
        if private_key is None:
            w3account, mnemonic = eth_account.Account.create_with_mnemonic()
            print(f"mnemonic: '{bright_cyan}{mnemonic}{color}'")
        else:
            w3account = web3.eth.account.from_key(private_key)

        if w3account.address in self._accounts:
            return self.at(w3account.address)

        account = LocalAccount(w3account.address, w3account, w3account.key)
        self._accounts.append(account)

        return account

    def from_mnemonic(
        self, mnemonic: str, count: int = 1, offset: int = 0, passphrase: str = ""
    ) -> Union["LocalAccount", list["LocalAccount"]]:
        """
        Generate one or more `LocalAccount` objects from a seed phrase.

        Arguments
        ---------
        mnemonic : str
            Space-separated list of BIP39 mnemonic seed words
        count : int, optional
            The number of `LocalAccount` objects to create
        offset : int, optional
            The initial account index to create accounts from
        passphrase : str, optional
            Additional passphrase to combine with the mnemonnic
        """
        new_accounts = []

        for i in range(offset, offset + count):
            w3account = eth_account.Account.from_mnemonic(
                mnemonic, passphrase=passphrase, account_path=f"m/44'/60'/0'/0/{i}"
            )

            account = LocalAccount(w3account.address, w3account, w3account.key)
            new_accounts.append(account)
            if account not in self._accounts:
                self._accounts.append(account)

        if count == 1:
            return new_accounts[0]
        return new_accounts

    def load(
        self,
        filename: str | Path | None = None,
        password: str | None = None,
        allow_retry: bool = False,
    ) -> Union[list, "LocalAccount"]:
        """
        Load a local account from a keystore file.

        Arguments
        ---------
        filename: str
            Keystore filename. If `None`, returns a list of available keystores.
        password: str
            Password to unlock the keystore. If `None`, password is entered via
            a getpass prompt.
        allow_retry: bool
            If True, allows re-attempt when the given password is incorrect.

        Returns
        -------
        LocalAccount
        """
        base_accounts_path = _get_data_folder().joinpath("accounts")
        if not filename:
            return [i.stem for i in base_accounts_path.glob("*.json")]

        filename = str(filename)
        json_file = Path(filename).expanduser()

        if not json_file.exists() or json_file.is_dir():
            temp_json_file = json_file.with_suffix(".json")
            if temp_json_file.exists():
                json_file = temp_json_file
            else:
                json_file_name = json_file.name
                json_file = base_accounts_path.joinpath(json_file_name)
                if json_file.suffix != ".json":
                    json_file = json_file.with_suffix(".json")
                if not json_file.exists():
                    raise FileNotFoundError(f"Cannot find {json_file}")

        with json_file.open() as fp:
            encrypted = ujson_load(fp)

        prompt = f'Enter password for "{json_file.stem}": '
        while True:
            if password is None:
                password = getpass(prompt)
            try:
                priv_key = web3.eth.account.decrypt(encrypted, password)
                break
            except ValueError as e:
                if allow_retry:
                    prompt = "Incorrect password, try again: "
                    password = None
                    continue
                raise e

        return self.add(priv_key)

    def at(self, address: str, force: bool = False) -> "LocalAccount":
        """
        Retrieve an `Account` instance from the address string.

        Raises `ValueError` if the account cannot be found.

        Arguments
        ---------
        address: string
            Address of the account
        force: bool
            When True, will add the account even if it was not found in web3.eth.accounts

        Returns
        -------
        Account
        """
        address = _resolve_address(address)
        acct = next((i for i in self._accounts if i == address), None)

        if acct is None and (address in web3.eth.accounts or force):
            acct = Account(address)

            if CONFIG.network_type == "development" and address not in web3.eth.accounts:
                rpc.unlock_account(address)

            self._accounts.append(acct)

        if acct:
            return acct
        raise UnknownAccount(f"No account exists for {address}")

    def remove(self, address: Union[str, "Account"]) -> None:
        """
        Remove an `Account` instance from the container.

        Arguments
        ---------
        address: str | Account
            Account instance, or string of the account address.
        """
        address = _resolve_address(str(address))
        try:
            self._accounts.remove(address)
        except ValueError:
            raise UnknownAccount(f"No account exists for {address}")

    def clear(self) -> None:
        """
        Empty the container.
        """
        self._accounts.clear()

    def connect_to_clef(self, uri: str = None, timeout: int = 120) -> None:
        """
        Connect to Clef and import open accounts.

        Clef is an account signing utility packaged with Geth, which can be
        used to interact with HW wallets in Brownie. Before calling this
        function, Clef must be running in another command prompt.

        Arguments
        ---------
        uri : str
            IPC path or http url to use to connect to clef. If None is given,
            uses the default IPC path on Unix systems or localhost on Windows.
        timeout : int
            The number of seconds to wait on a clef request before raising a
            timeout exception.
        """
        provider = None
        if uri is None:
            if sys.platform == "win32":
                uri = "http://localhost:8550/"
            else:
                uri = Path.home().joinpath(".clef/clef.ipc").as_posix()
        try:
            if Path(uri).exists():
                provider = IPCProvider(uri, timeout=timeout)
        except OSError:
            if uri is not None and uri.startswith("http"):
                provider = HTTPProvider(uri, {"timeout": timeout})
        if provider is None:
            raise ValueError("Unknown URI, must be IPC socket path or URL starting with 'http'")

        response = provider.make_request("account_list", [])
        if "error" in response:
            raise ValueError(response["error"]["message"])

        for address in response["result"]:
            if to_address(address) not in self._accounts:
                self._accounts.append(ClefAccount(address, provider))

    def disconnect_from_clef(self) -> None:
        """
        Disconnect from Clef.

        Removes all `ClefAccount` objects from the container.
        """
        self._accounts = [i for i in self._accounts if not isinstance(i, ClefAccount)]


class PublicKeyAccount:
    """Class for interacting with an Ethereum account where you do not control
    the private key. Can be used to check the balance or nonce, and to send ether to."""

    def __init__(self, addr: str) -> None:
        self.address = _resolve_address(addr)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} '{self.address}'>"

    def __hash__(self) -> int:
        return hash(self.address)

    def __str__(self) -> str:
        return self.address

    def __eq__(self, other: object | str) -> bool:
        if isinstance(other, str):
            try:
                address = _resolve_address(other)
                return address == self.address
            except ValueError:
                return False
        if isinstance(other, PublicKeyAccount):
            return other.address == self.address
        return super().__eq__(other)

    def balance(self) -> Wei:
        """Returns the current balance at the address, in wei."""
        balance = web3.eth.get_balance(self.address)
        return Wei(balance)

    @property
    def gas_used(self) -> int:
        """Returns the cumulative gas amount paid from this account."""
        return sum(i.gas_used for i in history.from_sender(self.address))

    @property
    def nonce(self) -> int:
        return web3.eth.get_transaction_count(self.address)

    def get_deployment_address(self, nonce: int | None = None) -> EthAddress:
        """
        Return the address of a contract deployed from this account at the given nonce.

        Arguments
        ---------
        nonce : int, optional
            The nonce of the deployment transaction. If not given, the nonce of the next
            transaction is used.
        """
        if nonce is None:
            nonce = self.nonce

        address = HexBytes(self.address)
        raw = rlp.encode([address, nonce])
        deployment_address = keccak(raw)[12:]

        return EthAddress(deployment_address)


class _PrivateKeyAccount(PublicKeyAccount):
    """Base class for Account and LocalAccount"""

    def __init__(self, addr: str) -> None:
        self._lock = threading.Lock()
        super().__init__(addr)

    def _pending_nonce(self) -> int:
        tx_from_sender = sorted(history.from_sender(self.address), key=lambda k: k.nonce)
        if len(tx_from_sender) == 0:
            return self.nonce

        last_tx = tx_from_sender[-1]
        if last_tx.status == -1:
            return last_tx.nonce + 1

        nonce = self.nonce
        while nonce == last_tx.nonce:
            # ganache does not always immediately increment the nonce
            time.sleep(0.5)
            nonce = self.nonce

        return nonce

    def _gas_limit(
        self,
        to: Optional["Account"],
        amount: int,
        gas_price: int | None,
        gas_buffer: float | None,
        data: str | None = None,
    ) -> int:
        gas_limit = CONFIG.active_network["settings"]["gas_limit"]
        if gas_limit == "max":
            return Chain().block_gas_limit

        if isinstance(gas_limit, bool) or gas_limit in (None, "auto"):
            gas_buffer = gas_buffer or CONFIG.active_network["settings"]["gas_buffer"]
            gas_limit = self.estimate_gas(to, amount, 0, data or "")
            if gas_limit > 21000 and gas_buffer != 1:
                gas_limit = Wei(gas_limit * gas_buffer)
                return min(gas_limit, Chain().block_gas_limit)

        return Wei(gas_limit)

    def _gas_price(self, gas_price: Any = None) -> tuple[Wei, GasABC | None, Iterator | None]:
        # returns the gas price, gas strategy object, and active gas strategy iterator
        if gas_price is None:
            gas_price = CONFIG.active_network["settings"]["gas_price"]

        if isinstance(gas_price, GasABC):
            value = gas_price.get_gas_price()
            if isinstance(value, Iterator):
                # if `get_gas_price` returns an iterator, this is a gas strategy
                # intended for rebroadcasting. we need to retain both the strategy
                # object and the active gas price iterator
                return Wei(next(value)), gas_price, value
            else:
                # for simple strategies, we can simply use the generated gas price
                return Wei(value), None, None

        if isinstance(gas_price, Wei):
            return gas_price, None, None

        if isinstance(gas_price, bool) or gas_price in (None, "auto"):
            return web3.eth.generate_gas_price(), None, None

        return Wei(gas_price), None, None

    def _check_for_revert(self, tx: dict) -> None:
        try:
            # remove gas price related values to avoid issues post-EIP1559
            # https://github.com/ethereum/go-ethereum/pull/23027
            skip_keys = {"gasPrice", "maxFeePerGas", "maxPriorityFeePerGas"}
            web3.eth.call({k: v for k, v in tx.items() if k not in skip_keys and v})
        except ValueError as exc:
            exc = VirtualMachineError(exc)
            raise ValueError(
                f"Execution reverted during call: '{exc.revert_msg}'. This transaction will likely "
                "revert. If you wish to broadcast, include `allow_revert:True` as a parameter.",
            ) from None

    def deploy(
        self,
        contract: Any,
        *args: Any,
        amount: int = 0,
        gas_limit: int | None = None,
        gas_buffer: float | None = None,
        gas_price: int | None = None,
        max_fee: int | None = None,
        priority_fee: int | None = None,
        nonce: int | None = None,
        required_confs: int = 1,
        allow_revert: bool = None,
        silent: bool = None,
        publish_source: bool = False,
    ) -> Any:
        """Deploys a contract.

        Args:
            contract: ContractContainer instance.
            *args: Constructor arguments. The last argument may optionally be
                   a dictionary of transaction values.

        Kwargs:
            amount: Amount of ether to send with transaction, in wei.
            gas_limit: Gas limit of the transaction.
            gas_buffer: Multiplier to apply to gas limit.
            gas_price: Gas price of legacy transaction.
            max_fee: Max fee per gas of dynamic fee transaction.
            priority_fee: Max priority fee per gas of dynamic fee transaction.
            nonce: Nonce to use for the transaction.

        Returns:
            * Contract instance if the transaction confirms and the contract exists
            * TransactionReceipt if the transaction is pending or reverts
        """
        data = contract.deploy.encode_input(*args)
        receipt, exc = self._make_transaction(
            None,
            amount,
            gas_limit,
            gas_buffer,
            gas_price,
            max_fee,
            priority_fee,
            data,
            nonce,
            f"{contract._name}.constructor",
            required_confs,
            allow_revert,
            silent,
        )

        add_thread = threading.Thread(target=contract._add_from_tx, args=(receipt,), daemon=True)
        add_thread.start()

        if rpc.is_active():
            undo_thread = threading.Thread(
                target=Chain()._add_to_undo_buffer,
                args=(
                    receipt,
                    self.deploy,
                    (contract, *args),
                    {
                        "amount": amount,
                        "gas_limit": gas_limit,
                        "gas_buffer": gas_buffer,
                        "gas_price": gas_price,
                        "max_fee": max_fee,
                        "priority_fee": priority_fee,
                    },
                ),
                daemon=True,
            )
            undo_thread.start()

        if receipt.status != 1:
            receipt._raise_if_reverted(exc)
            return receipt

        add_thread.join()
        try:
            deployed_contract = contract.at(receipt.contract_address)
            if publish_source:
                contract.publish_source(deployed_contract, silent=silent)
            return deployed_contract
        except ContractNotFound:
            # if the contract self-destructed during deployment
            return receipt

    def estimate_gas(
        self, to: "Account" = None, amount: int = 0, gas_price: int = None, data: str = None
    ) -> int:
        """
        Estimate the gas cost for a transaction.

        Raises VirtualMachineError if the transaction would revert.

        Arguments
        ---------
        to : Account, optional
            Account instance or address string of transaction recipient.
        amount : int, optional
            Amount of ether to send in wei.
        gas_price : int, optional
            Gas price of the transaction.
        data : str, optional
            Transaction data hexstring.

        Returns
        -------
        Estimated gas value in wei.
        """
        tx: dict = {
            "from": self.address,
            "to": to_address(str(to)) if to else None,
            "value": Wei(amount),
            "data": HexBytes(data or ""),
        }
        if gas_price is not None:
            tx["gasPrice"] = web3.to_hex(gas_price)
        try:
            return web3.eth.estimate_gas(tx)
        except ValueError as exc:
            revert_gas_limit = CONFIG.active_network["settings"]["reverting_tx_gas_limit"]
            if revert_gas_limit == "max":
                revert_gas_limit = web3.eth.get_block("latest")["gasLimit"]
                CONFIG.active_network["settings"]["reverting_tx_gas_limit"] = revert_gas_limit
            if revert_gas_limit:
                return revert_gas_limit

            exc = VirtualMachineError(exc)
            raise ValueError(
                f"Gas estimation failed: '{exc.revert_msg}'. This transaction will likely revert. "
                "If you wish to broadcast, you must set the gas limit manually."
            )

    def transfer(
        self,
        to: "Account" = None,
        amount: int = 0,
        gas_limit: int | None = None,
        gas_buffer: float | None = None,
        gas_price: int | None = None,
        max_fee: int | None = None,
        priority_fee: int | None = None,
        data: str = None,
        nonce: int | None = None,
        required_confs: int = 1,
        allow_revert: bool = None,
        silent: bool = None,
    ) -> TransactionReceipt:
        """
        Broadcast a transaction from this account.

        Kwargs:
            to: Account instance or address string to transfer to.
            amount: Amount of ether to send, in wei.
            gas_limit: Gas limit of the transaction.
            gas_buffer: Multiplier to apply to gas limit.
            gas_price: Gas price of legacy transaction.
            max_fee: Max fee per gas of dynamic fee transaction.
            priority_fee: Max priority fee per gas of dynamic fee transaction.
            nonce: Nonce to use for the transaction.
            data: Hexstring of data to include in transaction.
            silent: Toggles console verbosity.

        Returns:
            TransactionReceipt object
        """

        receipt, exc = self._make_transaction(
            to,
            amount,
            gas_limit,
            gas_buffer,
            gas_price,
            max_fee,
            priority_fee,
            data or "",
            nonce,
            "",
            required_confs,
            allow_revert,
            silent,
        )

        if rpc.is_active():
            undo_thread = threading.Thread(
                target=Chain()._add_to_undo_buffer,
                args=(
                    receipt,
                    self.transfer,
                    (),
                    {
                        "to": to,
                        "amount": amount,
                        "gas_limit": gas_limit,
                        "gas_buffer": gas_buffer,
                        "gas_price": gas_price,
                        "max_fee": max_fee,
                        "priority_fee": priority_fee,
                        "data": data,
                    },
                ),
                daemon=True,
            )
            undo_thread.start()

        receipt._raise_if_reverted(exc)
        return receipt

    def _make_transaction(
        self,
        to: Optional["Account"],
        amount: int,
        gas_limit: int | None,
        gas_buffer: float | None,
        gas_price: int | None,
        max_fee: int | None,
        priority_fee: int | None,
        data: str,
        nonce: int | None,
        fn_name: str,
        required_confs: int,
        allow_revert: bool | None,
        silent: bool | None,
    ) -> tuple[TransactionReceipt, Exception | None]:
        # shared logic for `transfer` and `deploy`
        if gas_limit and gas_buffer:
            raise ValueError("Cannot set gas_limit and gas_buffer together")
        if silent is None:
            silent = bool(CONFIG.mode == "test" or CONFIG.argv["silent"])

        if gas_price is None:
            # if gas price is not explicitly set, load the default max fee and priority fee
            if max_fee is None:
                max_fee = CONFIG.active_network["settings"]["max_fee"] or None
            if priority_fee is None:
                priority_fee = CONFIG.active_network["settings"]["priority_fee"] or None

        if priority_fee == "auto":
            try:
                priority_fee = Chain().priority_fee
            except ValueError:
                # fallback to legacy transactions if network does not support EIP1559
                CONFIG.active_network["settings"]["priority_fee"] = None
                priority_fee = None

        try:
            # if max fee and priority fee are not set, use gas price
            if max_fee is None and priority_fee is None:
                gas_price, gas_strategy, gas_iter = self._gas_price(gas_price)
            else:
                gas_strategy, gas_iter = None, None
            gas_limit = Wei(gas_limit) or self._gas_limit(
                to, amount, gas_price or max_fee, gas_buffer, data
            )
        except ValueError as e:
            raise VirtualMachineError(e) from None

        with self._lock:
            # we use a lock here to prevent nonce issues when sending many tx's at once
            tx = {
                "from": self.address,
                "value": Wei(amount),
                "nonce": nonce if nonce is not None else self._pending_nonce(),
                "gas": web3.to_hex(gas_limit),
                "data": HexBytes(data),
            }
            if to:
                tx["to"] = to_address(str(to))
            tx = _apply_fee_to_tx(tx, gas_price, max_fee, priority_fee)
            txid = None
            while True:
                try:
                    response = self._transact(tx, allow_revert)
                    exc, revert_data = None, None
                    if txid is None:
                        txid = bytes_to_hexstring(response)
                        if not silent:
                            print(f"\rTransaction sent: {bright_blue}{txid}{color}")
                except ValueError as e:
                    if txid is None:
                        exc = VirtualMachineError(e)
                        if not hasattr(exc, "txid"):
                            raise exc from None
                        txid = exc.txid
                        print(f"\rTransaction sent: {bright_blue}{txid}{color}")
                        revert_data = (exc.revert_msg, exc.pc, exc.revert_type)
                try:
                    receipt = TransactionReceipt(
                        txid,
                        self,
                        silent=silent,
                        required_confs=required_confs,
                        is_blocking=False,
                        name=fn_name,
                        revert_data=revert_data,
                    )
                    break
                except (TransactionNotFound, ValueError):
                    if not silent:
                        sys.stdout.write(f"  Awaiting transaction in the mempool... {_marker[0]}\r")
                        sys.stdout.flush()
                        _marker.rotate(1)
                    time.sleep(1)

        receipt = self._await_confirmation(receipt, required_confs, gas_strategy, gas_iter)
        if receipt.status != 1 and exc is None:
            error_data = {
                "message": (
                    f"VM Exception while processing transaction: revert {receipt.revert_msg}"
                ),
                "code": -32000,
                "data": {
                    receipt.txid: {
                        "error": "revert",
                        "program_counter": receipt._revert_pc,
                        "return": receipt.return_value,
                        "reason": receipt.revert_msg,
                    },
                },
            }
            exc = VirtualMachineError(ValueError(error_data))

        return receipt, exc

    def _await_confirmation(
        self,
        receipt: TransactionReceipt,
        required_confs: int,
        gas_strategy: GasABC | None,
        gas_iter: Iterator | None,
    ) -> TransactionReceipt:
        # add to TxHistory before waiting for confirmation, this way the tx
        # object is available if the user exits blocking via keyboard interrupt
        history._add_tx(receipt)

        if gas_strategy is not None:
            gas_strategy.run(receipt, gas_iter)

        if required_confs == 0:
            # set 0-conf tx's as silent to hide the confirmation output
            receipt._silent = True
            return receipt

        try:
            receipt._confirmed.wait()
        except KeyboardInterrupt as exc:
            # set related transactions as silent
            receipt._silent = True
            for receipt in history.filter(
                sender=self, nonce=receipt.nonce, key=lambda k: k.status != -2
            ):
                receipt._silent = True
            raise exc.with_traceback(None)

        if receipt.status != -2:
            return receipt

        # if transaction was dropped (status -2), find and return the tx that confirmed
        replacements = history.filter(
            sender=self, nonce=receipt.nonce, key=lambda k: k.status != -2
        )
        while True:
            if not replacements:
                raise TransactionError(f"Tx dropped without known replacement: {receipt.txid}")
            if len(replacements) > 1:
                # in case we have multiple tx's where the status is still unresolved
                replacements = [i for i in replacements if i.status != 2]
                time.sleep(0.5)
            else:
                receipt = replacements[0]
                receipt._await_confirmation(required_confs=required_confs)
                return receipt


class Account(_PrivateKeyAccount):
    """Class for interacting with an Ethereum account.

    Attributes:
        address: Public address of the account.
        nonce: Current nonce of the account."""

    def _transact(self, tx: dict, allow_revert: bool) -> Any:
        if allow_revert is None:
            allow_revert = bool(CONFIG.network_type == "development")
        if not allow_revert:
            self._check_for_revert(tx)
        return web3.eth.send_transaction(tx)


class LocalAccount(_PrivateKeyAccount):
    """Class for interacting with an Ethereum account.

    Attributes:
        address: Public address of the account.
        nonce: Current nonce of the account.
        private_key: Account private key.
        public_key: Account public key."""

    def __init__(self, address: str, account: Account, priv_key: int | bytes | str) -> None:
        self._acct = account
        if not isinstance(priv_key, str):
            priv_key = bytes_to_hexstring(priv_key)
        self.private_key = priv_key
        self.public_key = eth_keys.keys.PrivateKey(HexBytes(priv_key)).public_key
        super().__init__(address)

    def save(self, filename: str, overwrite: bool = False, password: str | None = None) -> str:
        """Encrypts the private key and saves it in a keystore json.

        Attributes:
            filename: path to keystore file. If no folder is given, saved in
                      ~/.brownie/accounts
            overwrite: if True, will overwrite an existing file.
            password: Password used to encrypt the account. If none, you will be prompted

        Returns the absolute path to the keystore file as a string.
        """
        path = _get_data_folder().joinpath("accounts")
        path.mkdir(exist_ok=True)
        filename = str(filename)
        if not filename.endswith(".json"):
            filename += ".json"
        if not any(i in r"\/" for i in filename):
            json_file = path.joinpath(filename).resolve()
        else:
            json_file = Path(filename).expanduser().resolve()
        if not overwrite and json_file.exists():
            raise FileExistsError("Account with this identifier already exists")

        if password is None:
            password = getpass("Enter the password to encrypt this account with: ")

        encrypted = web3.eth.account.encrypt(self.private_key, password)
        encrypted["address"] = encrypted["address"].lower()
        with json_file.open("w") as fp:
            ujson_dump(encrypted, fp)
        return str(json_file)

    def sign_defunct_message(self, message: str) -> SignedMessage:
        """Signs an `EIP-191` using this account's private key.

        Args:
            message: An text

        Returns:
            An eth_account `SignedMessage` instance.
        """
        msg_hash_bytes = defunct_hash_message(text=message)
        eth_private_key = eth_keys.keys.PrivateKey(HexBytes(self.private_key))
        (v, r, s, eth_signature_bytes) = sign_message_hash(eth_private_key, msg_hash_bytes)
        if ETH_ACCOUNT_LT_0_13_0:
            return SignedMessage(
                messageHash=msg_hash_bytes,
                r=r,
                s=s,
                v=v,
                signature=HexBytes(eth_signature_bytes),
            )
        else:
            return SignedMessage(
                message_hash=msg_hash_bytes,
                r=r,
                s=s,
                v=v,
                signature=HexBytes(eth_signature_bytes),
            )

    def sign_message(self, message: EIP712Message) -> SignedMessage:
        """Signs an `EIP712Message` using this account's private key.

        Args:
            message: An `EIP712Message` instance.

        Returns:
            An eth_account `SignedMessage` instance.
        """
        # some of this code is from:
        # https://github.com/ethereum/eth-account/blob/00e7b10/eth_account/account.py#L577
        # https://github.com/ethereum/eth-account/blob/00e7b10/eth_account/account.py#L502
        msg_hash_bytes = HexBytes(_hash_eip191_message(message.signable_message))
        assert len(msg_hash_bytes) == 32, "The message hash must be exactly 32-bytes"
        eth_private_key = eth_keys.keys.PrivateKey(HexBytes(self.private_key))
        (v, r, s, eth_signature_bytes) = sign_message_hash(eth_private_key, msg_hash_bytes)
        if ETH_ACCOUNT_LT_0_13_0:
            return SignedMessage(
                messageHash=msg_hash_bytes,
                r=r,
                s=s,
                v=v,
                signature=HexBytes(eth_signature_bytes),
            )
        else:
            return SignedMessage(
                message_hash=msg_hash_bytes,
                r=r,
                s=s,
                v=v,
                signature=HexBytes(eth_signature_bytes),
            )

    def _transact(self, tx: dict, allow_revert: bool) -> None:
        if allow_revert is None:
            allow_revert = bool(CONFIG.network_type == "development")
        if not allow_revert:
            self._check_for_revert(tx)
        tx["chainId"] = web3.chain_id
        signed = self._acct.sign_transaction(tx)
        return web3.eth.send_raw_transaction(
            signed.rawTransaction if ETH_ACCOUNT_LT_0_13_0 else signed.raw_transaction
        )


class ClefAccount(_PrivateKeyAccount):
    """
    Class for interacting with an Ethereum account where signing is handled in Clef.
    """

    def __init__(self, address: str, provider: HTTPProvider | IPCProvider) -> None:
        self._provider = provider
        super().__init__(address)

    def _transact(self, tx: dict, allow_revert: bool) -> None:
        if allow_revert is None:
            allow_revert = bool(CONFIG.network_type == "development")
        if not allow_revert:
            self._check_for_revert(tx)

        formatters = {
            "nonce": web3.to_hex,
            "value": web3.to_hex,
            "chainId": web3.to_hex,
            "data": web3.to_hex,
            "from": to_address,
        }
        if "to" in tx:
            formatters["to"] = to_address

        tx["chainId"] = web3.chain_id
        tx = apply_formatters_to_dict(formatters, tx)

        response = self._provider.make_request("account_signTransaction", [tx])
        if "error" in response:
            raise ValueError(response["error"]["message"])
        return web3.eth.send_raw_transaction(response["result"]["raw"])


def _apply_fee_to_tx(
    tx: dict,
    gas_price: int | None = None,
    max_fee: int | None = None,
    priority_fee: int | None = None,
) -> dict:
    tx = tx.copy()

    if gas_price is not None:
        if max_fee or priority_fee:
            raise ValueError("gas_price and (max_fee, priority_fee) are mutually exclusive")
        tx["gasPrice"] = web3.to_hex(gas_price)
        return tx

    if priority_fee is None:
        raise InvalidTransaction("priority_fee must be defined")
    priority_fee = Wei(priority_fee)

    # no max_fee specified, infer from base_fee
    if max_fee is None:
        base_fee = Chain().base_fee
        max_fee = base_fee * 2 + priority_fee
    else:
        max_fee = Wei(max_fee)

    if priority_fee > max_fee:
        raise InvalidTransaction("priority_fee must not exceed max_fee")

    tx["maxFeePerGas"] = web3.to_hex(max_fee)
    tx["maxPriorityFeePerGas"] = web3.to_hex(priority_fee)
    tx["type"] = web3.to_hex(2)
    return tx



#!/usr/bin/python3
# mypy: disable-error-code="union-attr"

import os
import pathlib
import shutil
import sys
import warnings
import zipfile
from base64 import b64encode
from collections.abc import Iterator, KeysView
from io import BytesIO
from types import ModuleType
from typing import Any, Final, Literal
from urllib.parse import quote

import requests
import yaml
from eth_typing import ChecksumAddress, HexStr
from mypy_extensions import mypyc_attr
from solcx.exceptions import SolcNotInstalled
from tqdm import tqdm
from ujson import JSONDecodeError
from vvm.exceptions import VyperNotInstalled

from brownie._c_constants import (
    Path,
    Version,
    import_module,
    mapcat,
    regex_match,
    sha1,
    ujson_dump,
    ujson_load,
)
from brownie._config import (
    CONFIG,
    REQUEST_HEADERS,
    _get_data_folder,
    _load_project_compiler_config,
    _load_project_config,
    _load_project_dependencies,
    _load_project_envvars,
    _load_project_structure_config,
)
from brownie._expansion import expand_posix_vars
from brownie.exceptions import (
    BadProjectName,
    BrownieEnvironmentWarning,
    InvalidPackage,
    PragmaError,
    ProjectAlreadyLoaded,
    ProjectNotFound,
)
from brownie.network.contract import (
    Contract,
    ContractContainer,
    InterfaceContainer,
    ProjectContract,
)
from brownie.network.state import _add_contract, _remove_contract, _revert_register
from brownie.project import compiler
from brownie.project.build import BUILD_KEYS, INTERFACE_KEYS, Build
from brownie.project.sources import Sources, get_pragma_spec
from brownie.typing import (
    BuildJson,
    CompilerConfig,
    ContractBuildJson,
    ContractName,
    EvmVersion,
    InterfaceBuildJson,
    Language,
    SolcConfig,
    VyperConfig,
)
from brownie.utils import notify

BUILD_FOLDERS: Final = "contracts", "deployments", "interfaces"
MIXES_URL: Final = "https://github.com/brownie-mix/{}-mix/archive/{}.zip"

GITIGNORE: Final = """__pycache__
.env
.history
.hypothesis/
build/
reports/
"""

GITATTRIBUTES: Final = """*.sol linguist-language=Solidity
*.vy linguist-language=Python
"""

NamespaceId = ContractName | Literal["interface"]
ChainDeployments = dict[ContractName, list[ChecksumAddress]]
DeploymentMap = dict[int | str, ChainDeployments]

_loaded_projects: Final[list["Project"]] = []


# TODO: remove this decorator once weakref support is implemented
@mypyc_attr(native_class=False)
class _ProjectBase:

    _path: pathlib.Path | None
    _build_path: pathlib.Path | None
    _sources: Sources
    _build: Build
    _containers: dict[ContractName, ContractContainer]

    def _compile(
        self, contract_sources: dict, compiler_config: CompilerConfig, silent: bool
    ) -> None:
        compiler_config.setdefault("solc", {})

        allow_paths = None
        cwd = os.getcwd()
        path = self._path
        if path is not None:
            _install_dependencies(path)
            allow_paths = path.as_posix()
            os.chdir(path)

        try:
            solc_config = compiler_config["solc"]
            vyper_config = compiler_config["vyper"]

            project_evm_version = compiler_config["evm_version"]
            evm_version: dict[Language, EvmVersion | None] = {
                "Solidity": solc_config.get("evm_version", project_evm_version),
                "Vyper": vyper_config.get("evm_version", project_evm_version),
            }

            build_json = compiler.compile_and_format(
                contract_sources,
                solc_version=solc_config.get("version", None),
                vyper_version=vyper_config.get("version", None),
                optimize=solc_config.get("optimize", None),
                runs=solc_config.get("runs", None),
                evm_version=evm_version,
                silent=silent,
                allow_paths=allow_paths,
                remappings=solc_config.get("remappings", []),
                optimizer=solc_config.get("optimizer", None),
                viaIR=solc_config.get("viaIR", None),
            )
        finally:
            os.chdir(cwd)

        build = self._build
        build_path = self._build_path
        for alias, data in build_json.items():
            if build_path is not None and not data["sourcePath"].startswith("interface"):
                # interfaces should generate artifact in /build/interfaces/ not /build/contracts/
                if alias == data["contractName"]:
                    # if the alias == contract name, this is a part of the core project
                    path = build_path.joinpath(f"contracts/{alias}.json")
                else:
                    # otherwise, this is an artifact from an external dependency
                    path = build_path.joinpath(f"contracts/dependencies/{alias}.json")
                    for parent in list(path.parents)[::-1]:
                        parent.mkdir(exist_ok=True)
                with path.open("w") as fp:
                    ujson_dump(data, fp, sort_keys=True, indent=2, default=sorted)

            if alias == data["contractName"]:
                # only add artifacts from the core project for now
                build._add_contract(data)

    def _create_containers(self) -> None:
        # create container objects
        self.interface = InterfaceContainer(self)
        self._containers = {}

        for key, data in self._build.items():
            if data["type"] == "interface":
                self.interface._add(data["contractName"], data["abi"])
            if data.get("bytecode"):
                container = ContractContainer(self, data)  # type: ignore [arg-type]
                self._containers[key] = container
                setattr(self, container._name, container)

    def __getitem__(self, key: ContractName) -> ContractContainer:
        return self._containers[key]

    def __iter__(self) -> Iterator[ContractContainer]:
        for i in sorted(self._containers):
            yield self._containers[i]

    def __len__(self) -> int:
        return len(self._containers)

    def __contains__(self, item: ContractName) -> bool:
        return item in self._containers

    def dict(self) -> dict[ContractName, ContractContainer]:
        return dict(self._containers)

    def keys(self) -> KeysView[ContractName]:
        return self._containers.keys()


# TODO: remove this decorator once weakref support is implemented
@mypyc_attr(native_class=False)
class Project(_ProjectBase):
    """
    Top level dict-like container that holds data and objects related to
    a brownie project.

    Attributes:
        _path: Path object, absolute path to the project
        _name: Name that the project is loaded as
        _sources: project Source object
        _build: project Build object
    """

    _compiler_config: CompilerConfig

    def __init__(self, name: str, project_path: pathlib.Path, compile: bool = True) -> None:
        self._path = project_path
        self._envvars: Final = _load_project_envvars(project_path)
        self._structure: Final = expand_posix_vars(
            _load_project_structure_config(project_path), self._envvars
        )
        self._build_path = project_path.joinpath(self._structure["build"])

        self._name: Final = name
        self._active: bool = False
        self.load(compile=compile)

    def load(self, raise_if_loaded: bool = True, compile: bool = True) -> None:
        """Compiles the project contracts, creates ContractContainer objects and
        populates the namespace."""
        if self._active:
            if raise_if_loaded:
                raise ProjectAlreadyLoaded("Project is already active")
            return None

        project_path: pathlib.Path = self._path  # type: ignore [assignment]
        structure = self._structure
        contract_sources = _load_sources(project_path, structure["contracts"], False)
        interface_sources = _load_sources(project_path, structure["interfaces"], True)
        sources = Sources(contract_sources, interface_sources)
        self._sources = sources

        build = Build(self._sources)
        self._build = build

        contract_list = sources.get_contract_list()
        potential_dependencies: list[tuple[pathlib.Path, ContractBuildJson]] = []
        build_path = self._build_path

        for path in build_path.glob("contracts/*.json"):
            contract_build_json = _load_contract_build_json_from_disk(path)
            if not set(BUILD_KEYS).issubset(contract_build_json):
                path.unlink()
                continue
            if path.stem not in contract_list:
                potential_dependencies.append((path, contract_build_json))
                continue
            if isinstance(contract_build_json["allSourcePaths"], list):
                # this handles the format change in v1.7.0, it can be removed in a future release
                path.unlink()
                test_path = build_path.joinpath("tests.json")
                if test_path.exists():
                    test_path.unlink()
                continue
            if not project_path.joinpath(contract_build_json["sourcePath"]).exists():
                path.unlink()
                continue
            build._add_contract(contract_build_json)

        for path, contract_build_json in potential_dependencies:
            dependents = build.get_dependents(path.stem)  # type: ignore [arg-type]
            is_dependency = len(set(dependents) & set(contract_list)) > 0
            if is_dependency:
                build._add_contract(contract_build_json)
            else:
                path.unlink()

        interface_hashes: dict[str, HexStr] = {}
        interface_list = sources.get_interface_list()

        for path in build_path.glob("interfaces/*.json"):
            interface_build_json = _load_interface_build_json_from_disk(path)
            if (
                not set(INTERFACE_KEYS).issubset(interface_build_json)
                or path.stem not in interface_list
            ):
                path.unlink()
                continue
            build._add_interface(interface_build_json)
            interface_hashes[path.stem] = interface_build_json["sha1"]

        if compile:
            self._compiler_config = config = expand_posix_vars(
                _load_project_compiler_config(project_path), self._envvars
            )

            # compile updated sources, update build
            changed = self._get_changed_contracts(interface_hashes)
            self._compile(changed, config, False)
            self._compile_interfaces(interface_hashes)
        self._load_dependency_artifacts()

        self._create_containers()
        self._load_deployments()

        # add project to namespaces, apply import blackmagic
        name = self._name
        self.__all__: list[NamespaceId] = list(self._containers) + ["interface"]
        sys.modules[f"brownie.project.{name}"] = self  # type: ignore [assignment]
        sys.modules["brownie.project"].__dict__[name] = self
        sys.modules["brownie.project"].__all__.append(name)
        sys.modules["brownie.project"].__console_dir__.append(name)
        self._namespaces: list[dict[NamespaceId, ContractContainer | InterfaceContainer]] = [
            sys.modules["__main__"].__dict__,  # type: ignore [list-item]
            sys.modules["brownie.project"].__dict__,  # type: ignore [list-item]
        ]

        # register project for revert and reset
        _revert_register(self)

        self._active = True
        _loaded_projects.append(self)

    def _get_changed_contracts(self, compiled_hashes: dict[str, HexStr]) -> dict[str, str]:
        # get list of changed interfaces and contracts
        sources = self._sources
        new_hashes = sources.get_interface_hashes()

        # remove outdated build artifacts
        build = self._build
        for name in new_hashes:
            if compiled_hashes.get(name) != new_hashes[name]:
                build._remove_interface(name)

        contracts = {c for c in sources.get_contract_list() if self._compare_build_json(c)}
        for contract in list(contracts):
            dependents = build.get_dependents(contract)
            contracts.update(dependents)

        # remove outdated build artifacts
        for name in contracts:
            build._remove_contract(name)

        # get final list of changed source paths
        changed_set = list({sources.get_source_path(contract) for contract in contracts})
        return dict(zip(changed_set, (sources.get(changed) for changed in changed_set)))

    def _compare_build_json(self, contract_name: ContractName) -> bool:
        config = self._compiler_config
        # confirm that this contract was previously compiled
        try:
            source = self._sources.get(contract_name)
            build_json: ContractBuildJson = self._build.get(contract_name)  # type: ignore [assignment]
        except KeyError:
            return True
        # compare source hashes
        if build_json["sha1"] != sha1(source.encode()).hexdigest():
            return True
        # compare compiler settings
        compiler = build_json["compiler"]
        if build_json["language"] == "Solidity":
            # compare solc-specific compiler settings
            solc_config = config["solc"].copy()
            solc_config["remappings"] = None
            if not _solidity_compiler_equal(solc_config, compiler):
                return True
            # compare solc pragma against compiled version
            if Version(compiler["version"]) not in get_pragma_spec(source):
                return True
        else:
            if not _vyper_compiler_equal(config["vyper"], compiler):
                return True
        return False

    def _compile_interfaces(self, compiled_hashes: dict) -> None:
        sources = self._sources
        new_hashes = sources.get_interface_hashes()
        changed_paths = [
            sources.get_source_path(k, True)
            for k, v in new_hashes.items()
            if compiled_hashes.get(k) != v
        ]
        if not changed_paths:
            return

        print("Generating interface ABIs...")
        solc_config = self._compiler_config["solc"]
        changed_sources = {i: sources.get(i) for i in changed_paths}
        abi_json = compiler.get_abi(
            changed_sources,
            solc_version=solc_config.get("version", None),
            allow_paths=self._path.as_posix(),
            remappings=solc_config.get("remappings", []),
        )

        build = self._build
        build_path = self._build_path
        for name, abi in abi_json.items():
            with build_path.joinpath(f"interfaces/{name}.json").open("w") as fp:
                ujson_dump(abi, fp, sort_keys=True, indent=2, default=sorted)
            build._add_interface(abi)

    def _load_dependency_artifacts(self) -> None:
        build = self._build
        dep_build_path = self._build_path.joinpath("contracts/dependencies/")
        for path in list(dep_build_path.glob("**/*.json")):
            contract_alias = path.relative_to(dep_build_path).with_suffix("").as_posix()
            if build.get_dependents(contract_alias):
                with path.open() as fp:
                    build_json = ujson_load(fp)
                # json.load turns arrays into python lists but for "offset" we need tuples
                build_json["offset"] = tuple(build_json["offset"])
                pc_map: dict[Any, dict] = build_json["pcMap"]
                for counter in pc_map.values():
                    if "offset" in counter:
                        counter["offset"] = tuple(counter["offset"])
                build._add_contract(build_json, contract_alias)
            else:
                path.unlink()

    def _load_deployments(self) -> None:
        if CONFIG.network_type != "live" and not CONFIG.settings["dev_deployment_artifacts"]:
            return
        chainid = CONFIG.active_network["chainid"] if CONFIG.network_type == "live" else "dev"
        path = self._build_path.joinpath(f"deployments/{chainid}")
        path.mkdir(exist_ok=True)
        deployments = list(path.glob("*.json"))
        deployments.sort(key=lambda k: k.stat().st_mtime)
        deployment_map = self._load_deployment_map()

        build: BuildJson
        for build_json in deployments:
            with build_json.open() as fp:
                build = ujson_load(fp)

            contract_name = build["contractName"]
            if contract_name not in self._containers:
                build_json.unlink()
                continue
            if "pcMap" in build:
                contract = ProjectContract(self, build, build_json.stem)
            else:
                contract = Contract.from_abi(  # type: ignore [assignment]
                    contract_name, build_json.stem, build["abi"]
                )
                contract._project = self
            container = self._containers[contract_name]
            _add_contract(contract)
            container._contracts.append(contract)

            # update deployment map for the current chain
            instances = deployment_map.setdefault(chainid, {}).setdefault(contract_name, [])
            if build_json.stem in instances:
                instances.remove(build_json.stem)
            instances.insert(0, build_json.stem)

        self._save_deployment_map(deployment_map)

    def _load_deployment_map(self) -> DeploymentMap:
        deployment_map = {}
        map_path = self._build_path.joinpath("deployments/map.json")
        if map_path.exists():
            with map_path.open("r") as fp:
                deployment_map = ujson_load(fp)
        return deployment_map

    def _save_deployment_map(self, deployment_map: DeploymentMap) -> None:
        with self._build_path.joinpath("deployments/map.json").open("w") as fp:
            ujson_dump(deployment_map, fp, sort_keys=True, indent=2, default=sorted)

    def _remove_from_deployment_map(self, contract: ProjectContract) -> None:
        if CONFIG.network_type != "live" and not CONFIG.settings["dev_deployment_artifacts"]:
            return
        chainid = CONFIG.active_network["chainid"] if CONFIG.network_type == "live" else "dev"
        deployment_map = self._load_deployment_map()
        try:
            deployment_map[chainid][contract._name].remove(contract.address)
            if not deployment_map[chainid][contract._name]:
                del deployment_map[chainid][contract._name]
            if not deployment_map[chainid]:
                del deployment_map[chainid]
        except (KeyError, ValueError):
            pass

        self._save_deployment_map(deployment_map)

    def _add_to_deployment_map(self, contract: ProjectContract) -> None:
        if CONFIG.network_type != "live" and not CONFIG.settings["dev_deployment_artifacts"]:
            return

        chainid = CONFIG.active_network["chainid"] if CONFIG.network_type == "live" else "dev"
        deployment_map = self._load_deployment_map()
        try:
            deployment_map[chainid][contract._name].remove(contract.address)
        except (ValueError, KeyError):
            pass
        deployment_map.setdefault(chainid, {}).setdefault(contract._name, []).insert(
            0, contract.address
        )
        self._save_deployment_map(deployment_map)

    def _update_and_register(
        self,
        dict_: dict[NamespaceId, ContractContainer | InterfaceContainer],
    ) -> None:
        dict_.update(self)  # type: ignore [arg-type]
        if "interface" not in dict_:
            dict_["interface"] = self.interface
        self._namespaces.append(dict_)

    def _add_to_main_namespace(self) -> None:
        # temporarily adds project objects to the main namespace
        brownie: ModuleType = sys.modules["brownie"]
        if "interface" not in brownie.__dict__:
            brownie.__dict__["interface"] = self.interface
        brownie.__dict__.update(self._containers)  # type: ignore [arg-type]
        brownie.__all__.extend(self.__all__)

    def _remove_from_main_namespace(self) -> None:
        # removes project objects from the main namespace
        brownie: ModuleType = sys.modules["brownie"]
        if brownie.__dict__.get("interface") == self.interface:
            del brownie.__dict__["interface"]
        for key in self._containers:
            brownie.__dict__.pop(key, None)
        for key in self.__all__:  # type: ignore [assignment]
            if key in brownie.__all__:
                brownie.__all__.remove(key)

    def __repr__(self) -> str:
        return f"<Project '{self._name}'>"

    def load_config(self) -> None:
        """Loads the project config file settings"""
        if isinstance(self._path, Path):
            _load_project_config(self._path)

    def close(self, raises: bool = True) -> None:
        """Removes pointers to the project's ContractContainer objects and this object."""
        if not self._active:
            if not raises:
                return
            raise ProjectNotFound("Project is not currently loaded.")

        # remove objects from namespace
        for dict_ in self._namespaces:
            for k, v in dict_.copy().items():
                if v == self or (k in self and v == self[k]):  # type: ignore [operator, index]
                    del dict_[k]

        # remove contracts
        for container in self._containers.values():
            for contract in container._contracts:
                _remove_contract(contract)
            container._contracts.clear()
        self._containers.clear()

        # undo black-magic
        self._remove_from_main_namespace()
        name = self._name
        del sys.modules[f"brownie.project.{name}"]
        sys.modules["brownie.project"].__all__.remove(name)
        sys.modules["brownie.project"].__console_dir__.remove(name)
        self._active = False
        _loaded_projects.remove(self)

        # clear paths
        try:
            sys.path.remove(str(self._path))
        except ValueError:
            pass

    def _clear_dev_deployments(self, height: int) -> None:
        path = self._build_path.joinpath("deployments/dev")
        if path.exists():
            deployment_map = self._load_deployment_map()
            for deployment in path.glob("*.json"):
                if height == 0:
                    deployment.unlink()
                else:
                    with deployment.open("r") as fp:
                        deployment_artifact: dict = ujson_load(fp)
                    block_height = deployment_artifact["deployment"]["blockHeight"]
                    if block_height > height:
                        deployment.unlink()
                        address = deployment_artifact["deployment"]["address"]
                        contract_name = deployment_artifact["contractName"]
                        try:
                            deployment_map["dev"][contract_name].remove(address)
                        except (KeyError, ValueError):
                            pass
            if "dev" in deployment_map and (height == 0 or not deployment_map["dev"]):
                del deployment_map["dev"]
                shutil.rmtree(path)

            self._save_deployment_map(deployment_map)

    def _revert(self, height: int) -> None:
        self._clear_dev_deployments(height)

    def _reset(self) -> None:
        self._clear_dev_deployments(0)


def _load_contract_build_json_from_disk(path: pathlib.Path) -> ContractBuildJson:
    try:
        with path.open() as fp:
            contract_build_json: dict = ujson_load(fp)
            # json loads them as lists but we want tuples for mypyc
            contract_build_json["offset"] = tuple(contract_build_json["offset"])
            pc_map: dict = contract_build_json["pcMap"]
            counter: dict
            for counter in pc_map.values():
                if "offset" in counter:
                    counter["offset"] = tuple(counter["offset"])
            return contract_build_json  # type: ignore [return-value]
    except JSONDecodeError:
        return {}  # type: ignore [return-value]


def _load_interface_build_json_from_disk(path: pathlib.Path) -> InterfaceBuildJson:
    try:
        with path.open() as fp:
            interface_build_json: dict = ujson_load(fp)
            offset = interface_build_json["offset"]
            if offset is not None:
                interface_build_json["offset"] = tuple(offset)
            return interface_build_json  # type: ignore [return-value]
    except JSONDecodeError:
        return {}  # type: ignore [typeddict-item]


# TODO: remove this decorator once weakref support is implemented
@mypyc_attr(native_class=False)
class TempProject(_ProjectBase):
    """Simplified Project class used to hold temporary contracts that are
    compiled via project.compile_source"""

    def __init__(self, name: str, contract_sources: dict, compiler_config: CompilerConfig) -> None:
        self._path = None
        self._build_path = None
        self._name: Final = name
        self._sources = Sources(contract_sources, {})
        self._build = Build(self._sources)
        self._compile(contract_sources, compiler_config, True)
        self._create_containers()

    def __repr__(self) -> str:
        return f"<TempProject '{self._name}'>"


def check_for_project(path: pathlib.Path | str = ".") -> pathlib.Path | None:
    """Checks for a Brownie project."""
    path = Path(path).resolve()
    for folder in [path] + list(path.parents):

        structure_config = _load_project_structure_config(folder)
        contracts = folder.joinpath(structure_config["contracts"])
        interfaces = folder.joinpath(structure_config["interfaces"])
        scripts = folder.joinpath(structure_config["scripts"])
        tests = folder.joinpath(structure_config["tests"])

        if next((i for i in contracts.glob("**/*") if i.suffix in (".vy", ".sol")), None):
            return folder
        if next((i for i in interfaces.glob("**/*") if i.suffix in (".json", ".vy", ".sol")), None):
            return folder
        if next((i for i in scripts.glob("**/*") if i.suffix in (".py")), None):
            return folder
        if contracts.is_dir() and tests.is_dir():
            return folder

    return None


def get_loaded_projects() -> list["Project"]:
    """Returns a list of currently loaded Project objects."""
    return _loaded_projects.copy()


def new(
    project_path_str: str = ".", ignore_subfolder: bool = False, ignore_existing: bool = False
) -> str:
    """Initializes a new project.

    Args:
        project_path: Path to initialize the project at. If not exists, it will be created.
        ignore_subfolder: (deprecated)
        ignore_existing: If True, will not raise when initiating in a non-empty directory.

    Returns the path to the project as a string.
    """
    project_path = Path(project_path_str).resolve()
    if not ignore_existing and project_path.exists() and list(project_path.glob("*")):
        raise FileExistsError(f"Directory is not empty: {project_path}")
    project_path.mkdir(exist_ok=True)
    _create_folders(project_path)
    _create_gitfiles(project_path)
    _add_to_sys_path(project_path)
    return str(project_path)


def from_brownie_mix(
    project_name: str,
    project_path: pathlib.Path | str | None = None,
    ignore_subfolder: bool = False,
) -> str:
    """Initializes a new project via a template. Templates are downloaded from
    https://www.github.com/brownie-mix

    Args:
        project_path: Path to initialize the project at.
        ignore_subfolders: (deprecated)

    Returns the path to the project as a string.
    """
    project_name = str(project_name).lower().replace("-mix", "")
    headers = REQUEST_HEADERS.copy()
    headers.update(_maybe_retrieve_github_auth())
    default_branch = _get_mix_default_branch(project_name, headers)
    url = MIXES_URL.format(project_name, default_branch)
    if project_path is None:
        project_path = Path(".").joinpath(project_name)
    project_path = Path(project_path).resolve()
    if project_path.exists() and list(project_path.glob("*")):
        raise FileExistsError(f"Folder already exists - {project_path}")

    print(f"Downloading from {url}...")
    _stream_download(url, str(project_path.parent), headers)
    project_path.parent.joinpath(f"{project_name}-mix-{default_branch}").rename(project_path)
    _create_folders(project_path)
    _create_gitfiles(project_path)
    _add_to_sys_path(project_path)
    return str(project_path)


def compile_source(
    source: str,
    solc_version: str | None = None,
    vyper_version: str | None = None,
    optimize: bool = True,
    runs: int | None = 200,
    evm_version: EvmVersion | None = None,
) -> "TempProject":
    """
    Compile the given source code string and return a TempProject container with
    the ContractContainer instances.
    """
    compiler_config: CompilerConfig = {"evm_version": evm_version, "solc": {}, "vyper": {}}

    # if no compiler version was given, first try to find a Solidity pragma
    if solc_version is None and vyper_version is None:
        try:
            solc_version = compiler.solidity.find_best_solc_version(
                {"<stdin>": source}, install_needed=True, silent=False
            )
        except (PragmaError, SolcNotInstalled):
            pass

    if vyper_version is None:
        # if no vyper compiler version is given, try to compile using solidity
        compiler_config["solc"] = {
            "version": solc_version or str(compiler.solidity.get_version().truncate()),
            "optimize": bool(optimize),
            "runs": runs or 0,
        }
        try:
            return TempProject("TempSolcProject", {"<stdin>.sol": source}, compiler_config)
        except Exception as exc:
            # if compilation fails, raise when a solc version was given or we found a pragma
            if solc_version is not None:
                raise exc

    if vyper_version is None:
        # if no vyper compiler version was given, try to find a pragma
        try:
            vyper_version = compiler.vyper.find_best_vyper_version(
                {"<stdin>": source}, install_needed=True, silent=False
            )
        except (PragmaError, VyperNotInstalled):
            pass

    compiler_config["vyper"] = {"version": vyper_version or compiler.vyper.get_version()}
    try:
        return TempProject("TempVyperProject", {"<stdin>.vy": source}, compiler_config)
    except Exception as exc:
        if solc_version is None and vyper_version is None:
            raise PragmaError(
                "No compiler version specified, no pragma statement in the source, "
                "and compilation failed with both solc and vyper"
            ) from None
        raise exc


def load(
    project_path: pathlib.Path | str | None = None,
    name: str | None = None,
    raise_if_loaded: bool = True,
    compile: bool = True,
) -> "Project":
    """Loads a project and instantiates various related objects.

    Args:
        project_path: Path of the project to load. If None, will attempt to
                      locate a project using check_for_project()
        name: Name to assign to the project. If None, the name is generated
              from the name of the project folder

    Returns a Project object.
    """
    # checks
    if project_path is None:
        project_path = check_for_project(".")
        if project_path is not None and project_path != Path(".").absolute():
            warnings.warn(
                f"Loaded project has a root folder of '{project_path}' "
                "which is different from the current working directory",
                BrownieEnvironmentWarning,
            )
    else:
        project_path = Path(project_path)
        if project_path.resolve() != check_for_project(project_path):
            packages_path = _get_data_folder().joinpath("packages")
            if not project_path.is_absolute() and packages_path.joinpath(project_path).exists():
                project_path = packages_path.joinpath(project_path)
            else:
                project_path = None

    if project_path is None:
        raise ProjectNotFound("Could not find Brownie project")

    project_path = Path(project_path).resolve()
    if name is None:
        name = project_path.name
        if not name.lower().endswith("project"):
            name += " project"
        if not name[0].isalpha():
            raise BadProjectName("Project must start with an alphabetic character")
        name = "".join(i for i in name.title() if i.isalnum())

    for loaded_project in _loaded_projects:
        if loaded_project._name == name:
            if raise_if_loaded:
                raise ProjectAlreadyLoaded("There is already a project loaded with this name")
            return loaded_project

    # paths
    _create_folders(project_path)
    _add_to_sys_path(project_path)

    # load sources and build
    return Project(name, project_path, compile=compile)


def _install_dependencies(path: pathlib.Path) -> None:
    for package_id in _load_project_dependencies(path):
        try:
            install_package(package_id)
        except FileExistsError:
            pass


def install_package(package_id: str) -> str:
    """
    Install a package.

    Arguments
    ---------
    package_id : str
        Package ID

    Returns
    -------
    str
        ID of the installed package.
    """
    return _install_from_github(package_id)


def _maybe_retrieve_github_auth() -> dict[str, str]:
    """Returns appropriate github authorization headers.

    Otherwise returns an empty dict if no auth token is present.
    """
    if token := os.getenv("GITHUB_TOKEN"):
        auth = b64encode(token.encode()).decode()
        return {"Authorization": f"Basic {auth}"}
    return {}


def _install_from_github(package_id: str) -> str:
    try:
        path, version = package_id.split("@", 1)
        org, repo = path.split("/")
    except ValueError:
        raise ValueError(
            "Invalid package ID. Must be given as [ORG]/[REPO]@[VERSION]"
            "\ne.g. 'OpenZeppelin/openzeppelin-contracts@v2.5.0'"
        ) from None

    base_install_path = _get_data_folder().joinpath("packages")
    install_path = base_install_path.joinpath(f"{org}")
    install_path.mkdir(exist_ok=True)
    install_path = install_path.joinpath(f"{repo}@{version}")
    if install_path.exists():
        raise FileExistsError("Package is already installed")

    headers = REQUEST_HEADERS.copy()
    headers.update(_maybe_retrieve_github_auth())

    if regex_match(r"^[0-9a-f]+$", version):
        download_url = f"https://api.github.com/repos/{org}/{repo}/zipball/{version}"
    else:
        download_url = _get_download_url_from_tag(org, repo, version, headers)

    existing = list(install_path.parent.iterdir())

    # Some versions contain special characters and github api seems to display url without
    # encoding them.
    # It results in a ConnectionError exception because the actual download url is encoded.
    # In this case we try to sanitize the version in url and download again.
    try:
        _stream_download(download_url, str(install_path.parent), headers)
    except ConnectionError:
        download_url = (
            f"https://api.github.com/repos/{org}/{repo}/zipball/refs/tags/{quote(version)}"
        )
        _stream_download(download_url, str(install_path.parent), headers)

    installed = next(i for i in install_path.parent.iterdir() if i not in existing)
    shutil.move(installed, install_path)

    try:
        if not install_path.joinpath("brownie-config.yaml").exists():
            brownie_config: dict = {"project_structure": {}}

            contract_paths = {
                i.relative_to(install_path).parts[0]
                for i in mapcat(install_path.glob, ("**/*.sol", "**/*.vy"))
            }
            if not contract_paths:
                raise InvalidPackage(f"{package_id} does not contain any .sol or .vy files")
            if install_path.joinpath("contracts").is_dir():
                brownie_config["project_structure"]["contracts"] = "contracts"
            elif len(contract_paths) == 1:
                brownie_config["project_structure"]["contracts"] = contract_paths.pop()
            else:
                raise Exception(
                    f"{package_id} has no `contracts/` subdirectory, and "
                    "multiple directories containing source files"
                )

            with install_path.joinpath("brownie-config.yaml").open("w") as fp:
                yaml.dump(brownie_config, fp)

            Path.touch(install_path / ".env")

        project = load(install_path)
        project.close()
    except InvalidPackage:
        shutil.rmtree(install_path)
        raise
    except Exception as e:
        notify(
            "WARNING",
            f"Unable to compile {package_id} due to a {type(e).__name__} - you may still be able to"
            " import sources from the package, but will be unable to load the package directly.\n",
        )

    return f"{org}/{repo}@{version}"


def _get_download_url_from_tag(org: str, repo: str, version: str, headers: dict) -> str:
    response = requests.get(
        f"https://api.github.com/repos/{org}/{repo}/tags?per_page=100", headers=headers
    )
    status_code = response.status_code
    if status_code != 200:
        message = response.json()["message"]
        msg = f"Status {status_code} when getting package versions from Github: '{message}'"
        if status_code in {403, 404}:
            msg += (
                "\n\nMissing or forbidden.\n"
                "If this issue persists, generate a Github API token and store"
                " it as the environment variable `GITHUB_TOKEN`:\n"
                "https://github.blog/2013-05-16-personal-api-tokens/"
            )
        raise ConnectionError(msg)

    data = response.json()
    if not data:
        raise ValueError("Github repository has no tags set")
    org, repo = data[0]["zipball_url"].split("/")[3:5]
    tags = [i["name"].lstrip("v") for i in data]
    if version not in tags:
        raise ValueError(
            "Invalid version for this package. Available versions are:\n" + ", ".join(tags)
        ) from None

    return next(i["zipball_url"] for i in data if i["name"].lstrip("v") == version)


def _create_gitfiles(project_path: pathlib.Path) -> None:
    gitignore = project_path.joinpath(".gitignore")
    if not gitignore.exists():
        with gitignore.open("w") as fp:
            fp.write(GITIGNORE)
    gitattributes = project_path.joinpath(".gitattributes")
    if not gitattributes.exists():
        with gitattributes.open("w") as fp:
            fp.write(GITATTRIBUTES)


def _create_folders(project_path: pathlib.Path) -> None:
    structure = _load_project_structure_config(project_path)
    for path in structure.values():
        project_path.joinpath(path).mkdir(exist_ok=True)
    build_path = project_path.joinpath(structure["build"])
    for path in BUILD_FOLDERS:
        build_path.joinpath(path).mkdir(exist_ok=True)


def _add_to_sys_path(project_path: pathlib.Path) -> None:
    project_path_string = str(project_path)
    if project_path_string in sys.path:
        return
    sys.path.insert(0, project_path_string)


def _compare_settings(left: dict, right: dict) -> bool:
    return any(v and not isinstance(v, dict) and v != right.get(k) for k, v in left.items())


def _normalize_solidity_version(version: str) -> str:
    return version.split("+")[0]


def _solidity_compiler_equal(config: SolcConfig, build: CompilerConfig) -> bool:
    return (
        config["version"] is None
        or _normalize_solidity_version(config["version"])
        == _normalize_solidity_version(build["version"])
    ) and config["optimizer"] == build["optimizer"]


def _vyper_compiler_equal(config: VyperConfig, build: CompilerConfig) -> bool:
    return config["version"] is None or config["version"] == build["version"]


def _load_sources(project_path: pathlib.Path, subfolder: str, allow_json: bool) -> dict:
    contract_sources: dict = {}
    suffixes: tuple = (".sol", ".vy")
    if allow_json:
        suffixes = suffixes + (".json",)

    # one day this will be a beautiful plugin system
    hooks: ModuleType | None = None
    if project_path.joinpath("brownie_hooks.py").exists():
        hooks = import_module("brownie_hooks")

    for path in project_path.glob(f"{subfolder}/**/*"):
        if path.suffix not in suffixes:
            continue
        if next((i for i in path.relative_to(project_path).parts if i.startswith("_")), False):
            continue
        with path.open(encoding="utf-8") as fp:
            source = fp.read()

        if hasattr(hooks, "brownie_load_source"):
            source = hooks.brownie_load_source(path, source)

        path_str: str = path.relative_to(project_path).as_posix()
        contract_sources[path_str] = source
    return contract_sources


def _stream_download(
    download_url: str, target_path: str, headers: dict[str, str] = REQUEST_HEADERS
) -> None:
    response = requests.get(download_url, stream=True, headers=headers)

    if response.status_code == 404:
        raise ConnectionError(
            f"404 error when attempting to download from {download_url} - "
            "are you sure this is a valid mix? https://github.com/brownie-mix"
        )
    if response.status_code != 200:
        raise ConnectionError(
            f"Received status code {response.status_code} when attempting "
            f"to download from {download_url}"
        )

    total_size = int(response.headers.get("content-length", 0))
    progress_bar = tqdm(total=total_size, unit="iB", unit_scale=True)
    content = b""

    for data in response.iter_content(1024, decode_unicode=True):
        progress_bar.update(len(data))
        content += data
    progress_bar.close()

    with zipfile.ZipFile(BytesIO(content)) as zf:
        zf.extractall(target_path)


def _get_mix_default_branch(mix_name: str, headers: dict[str, str] = REQUEST_HEADERS) -> str:
    """Get the default branch for a brownie-mix repository.

    Arguments
    ---------
    mix_name : str
        Name of a brownie-mix repository without -mix appended.

    Returns
    -------
    str
        The default branch name on github.
    """
    REPO_GH_API = f"https://api.github.com/repos/brownie-mix/{mix_name}-mix"
    r = requests.get(REPO_GH_API, headers=headers)
    if r.status_code != 200:
        status, repo, message = r.status_code, f"brownie-mix/{mix_name}", r.json()["message"]
        msg = f"Status {status} when retrieving repo {repo} information from GHAPI: '{message}'"
        if r.status_code in {403, 404}:
            msg_lines = (
                msg,
                "\n\nMissing or forbidden.\n",
                "If this issue persists, generate a Github API token and store",
                " it as the environment variable `GITHUB_TOKEN`:\n",
                "https://github.blog/2013-05-16-personal-api-tokens/",
            )
            msg = "".join(msg_lines)
        raise ConnectionError(msg)
    elif "default_branch" not in r.json():
        msg = f"API results did not include {mix_name}'s default branch"
        raise KeyError(msg)

    return r.json()["default_branch"]



#!/usr/bin/python3

from typing import Final

from .account import Accounts
from .main import (  # NOQA 401
    connect,
    disconnect,
    gas_limit,
    gas_price,
    is_connected,
    max_fee,
    priority_fee,
    show_active,
)
from .rpc import Rpc
from .state import Chain, TxHistory
from .web3 import web3

__all__ = ["accounts", "chain", "history", "rpc", "web3"]
__console_dir__ = ["connect", "disconnect", "show_active", "is_connected", "gas_limit", "gas_price"]

accounts: Final[Accounts] = Accounts()
rpc: Final[Rpc] = Rpc()
history: Final[TxHistory] = TxHistory()
chain: Final[Chain] = Chain()
